from datetime import timedelta

import pytest
from django.utils import timezone

from ...conf.test import override_dynamic_settings
from ..api.postingendpoint import PostingEndpoint, PostingInterrupt
from ..api.postingendpoint.floodprotection import FloodProtectionMiddleware
from ..test import post_thread

default_acl = {"can_omit_flood_protection": False}
can_omit_flood_acl = {"can_omit_flood_protection": True}


def test_middleware_lets_users_first_post_through(dynamic_settings, user):
    user.update_fields = []
    middleware = FloodProtectionMiddleware(
        settings=dynamic_settings, user=user, user_acl=default_acl
    )
    middleware.interrupt_posting(None)


def test_middleware_updates_users_last_post_datetime(dynamic_settings, user):
    user.update_fields = []
    middleware = FloodProtectionMiddleware(
        settings=dynamic_settings, user=user, user_acl=default_acl
    )
    middleware.interrupt_posting(None)
    assert user.last_posted_on
    assert "last_posted_on" in user.update_fields


def test_middleware_interrupts_posting_because_previous_post_was_posted_too_recently(
    dynamic_settings, user
):
    user.last_posted_on = timezone.now()
    user.update_fields = []

    middleware = FloodProtectionMiddleware(
        mode=PostingEndpoint.START,
        settings=dynamic_settings,
        user=user,
        user_acl=default_acl,
    )
    assert middleware.use_this_middleware()

    with pytest.raises(PostingInterrupt):
        middleware.interrupt_posting(None)


def test_middleware_lets_users_next_post_through_if_previous_post_is_not_recent(
    dynamic_settings, user
):
    user.last_posted_on = timezone.now() - timedelta(seconds=10)
    user.update_fields = []

    middleware = FloodProtectionMiddleware(
        mode=PostingEndpoint.START,
        settings=dynamic_settings,
        user=user,
        user_acl=default_acl,
    )
    assert middleware.use_this_middleware()
    middleware.interrupt_posting(None)


def test_middleware_is_not_used_if_user_has_permission_to_omit_flood_protection(
    dynamic_settings, user
):
    user.last_posted_on = timezone.now()
    user.update_fields = []

    middleware = FloodProtectionMiddleware(
        mode=PostingEndpoint.START,
        settings=dynamic_settings,
        user=user,
        user_acl=can_omit_flood_acl,
    )
    assert not middleware.use_this_middleware()


def test_middleware_is_not_used_if_user_edits_post(dynamic_settings, user):
    middleware = FloodProtectionMiddleware(
        mode=PostingEndpoint.EDIT,
        settings=dynamic_settings,
        user=user,
        user_acl=can_omit_flood_acl,
    )
    assert not middleware.use_this_middleware()


@override_dynamic_settings(hourly_post_limit=3)
def test_middleware_interrupts_posting_if_hourly_limit_was_met(
    default_category, dynamic_settings, user
):
    user.update_fields = []

    for _ in range(3):
        post_thread(default_category, poster=user)

    middleware = FloodProtectionMiddleware(
        mode=PostingEndpoint.START,
        settings=dynamic_settings,
        user=user,
        user_acl=default_acl,
    )
    assert middleware.use_this_middleware()

    with pytest.raises(PostingInterrupt):
        middleware.interrupt_posting(None)


@override_dynamic_settings(hourly_post_limit=0)
def test_old_posts_dont_count_to_hourly_limit(default_category, dynamic_settings, user):
    user.update_fields = []

    for _ in range(3):
        post_thread(
            default_category,
            poster=user,
            started_on=timezone.now() - timedelta(hours=1),
        )

    middleware = FloodProtectionMiddleware(
        mode=PostingEndpoint.START,
        settings=dynamic_settings,
        user=user,
        user_acl=default_acl,
    )
    middleware.interrupt_posting(None)


@override_dynamic_settings(hourly_post_limit=0)
def test_middleware_lets_post_through_if_hourly_limit_was_disabled(
    default_category, dynamic_settings, user
):
    user.update_fields = []

    for _ in range(3):
        post_thread(default_category, poster=user)

    middleware = FloodProtectionMiddleware(
        mode=PostingEndpoint.START,
        settings=dynamic_settings,
        user=user,
        user_acl=default_acl,
    )
    middleware.interrupt_posting(None)


@override_dynamic_settings(daily_post_limit=3)
def test_middleware_interrupts_posting_if_daily_limit_was_met(
    default_category, dynamic_settings, user
):
    user.update_fields = []

    for _ in range(3):
        post_thread(default_category, poster=user)

    middleware = FloodProtectionMiddleware(
        mode=PostingEndpoint.START,
        settings=dynamic_settings,
        user=user,
        user_acl=default_acl,
    )
    assert middleware.use_this_middleware()

    with pytest.raises(PostingInterrupt):
        middleware.interrupt_posting(None)


@override_dynamic_settings(daily_post_limit=0)
def test_old_posts_dont_count_to_daily_limit(default_category, dynamic_settings, user):
    user.update_fields = []

    for _ in range(3):
        post_thread(
            default_category, poster=user, started_on=timezone.now() - timedelta(days=1)
        )

    middleware = FloodProtectionMiddleware(
        mode=PostingEndpoint.START,
        settings=dynamic_settings,
        user=user,
        user_acl=default_acl,
    )
    middleware.interrupt_posting(None)


@override_dynamic_settings(daily_post_limit=0)
def test_middleware_lets_post_through_if_daily_limit_was_disabled(
    default_category, dynamic_settings, user
):
    user.update_fields = []

    for _ in range(3):
        post_thread(default_category, poster=user)

    middleware = FloodProtectionMiddleware(
        mode=PostingEndpoint.START,
        settings=dynamic_settings,
        user=user,
        user_acl=default_acl,
    )
    middleware.interrupt_posting(None)



from django.contrib.auth import get_user_model
from rest_framework import serializers

from ...notifications.models import Notification
from ...notifications.registry import registry

User = get_user_model()


class NotificationActorSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = [
            "id",
            "username",
            "avatars",
        ]

    def to_representation(self, obj: User) -> dict:
        data = super().to_representation(obj)
        data["url"] = obj.get_absolute_url()
        return data


class NotificationSerializer(serializers.ModelSerializer):
    class Meta:
        model = Notification
        fields = ["created_at"]

    def to_representation(self, obj: Notification) -> dict:
        data = super().to_representation(obj)

        if obj.actor:
            actor_data = NotificationActorSerializer(obj.actor).data
        else:
            actor_data = None

        return {
            "id": obj.id,
            "isRead": obj.is_read,
            "createdAt": data["created_at"],
            "actor": actor_data,
            "actorName": obj.actor_name,
            "message": registry.get_message(obj),
            "url": obj.get_absolute_url(),
        }



from django.apps import AppConfig


class MisagoLegalConfig(AppConfig):
    name = "misago.legal"
    label = "misago_legal"
    verbose_name = "Misago Legal"

    def ready(self):
        from . import signals as _



from django import forms
from django.utils.translation import pgettext_lazy

from ..models import Role
from ..providers import providers


class RoleForm(forms.ModelForm):
    name = forms.CharField(label=pgettext_lazy("admin role form", "Role name"))

    class Meta:
        model = Role
        fields = ["name"]


def get_permissions_forms(role, data=None):
    """utility function for building forms in admin"""
    role_permissions = role.permissions

    perms_forms = []
    for extension, module in providers.list():
        try:
            module.change_permissions_form
        except AttributeError:
            message = "'%s' object has no attribute '%s'"
            raise AttributeError(message % (extension, "change_permissions_form"))

        FormType = module.change_permissions_form(role)

        if FormType:
            if data:
                form = FormType(data, prefix=extension)
            else:
                form = FormType(
                    initial=role_permissions.get(extension), prefix=extension
                )
            perms_forms.append(form)

    return perms_forms



from django.conf import settings
from django.db import models

from ..avatars import store


class Avatar(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    size = models.PositiveIntegerField(default=0)
    image = models.ImageField(max_length=255, upload_to=store.upload_to)

    @property
    def url(self):
        return self.image.url



import pytest

from ....acl.models import Role
from ....test import assert_contains, assert_not_contains
from ...models import Rank
from ...test import create_test_user


@pytest.fixture
def user_a(db):
    return create_test_user("Tyrael", "test123@test.org")


@pytest.fixture
def user_b(db):
    return create_test_user("Tyrion", "test321@gmail.com")


@pytest.fixture
def user_c(db):
    return create_test_user("Karen", "other432@gmail.com")


def test_search_finds_user_by_username(admin_client, users_admin_link, user_a):
    response = admin_client.get("%s&username=Tyrael" % users_admin_link)
    assert_contains(response, user_a.username)


def test_search_finds_user_by_username_with_underscore(
    admin_client, users_admin_link, other_user
):
    response = admin_client.get("%s&username=other_user" % users_admin_link)
    assert_contains(response, other_user.username)


def test_search_excludes_users_with_different_username(
    admin_client, users_admin_link, user_b, user_c
):
    response = admin_client.get("%s&username=Tyrael" % users_admin_link)
    assert_not_contains(response, user_b.username)
    assert_not_contains(response, user_c.username)


def test_search_finds_users_by_username_start(
    admin_client, users_admin_link, user_a, user_b
):
    response = admin_client.get("%s&username=Tyr*" % users_admin_link)
    assert_contains(response, user_a.username)
    assert_contains(response, user_b.username)


def test_search_finds_user_username_with_underscore_by_username_start(
    admin_client, users_admin_link, other_user
):
    response = admin_client.get("%s&username=other_*" % users_admin_link)
    assert_contains(response, other_user.username)


def test_search_by_username_start_excludes_users_with_different_username(
    admin_client, users_admin_link, user_c
):
    response = admin_client.get("%s&username=Tyr*" % users_admin_link)
    assert_not_contains(response, user_c.username)


def test_search_finds_user_by_username_end(admin_client, users_admin_link, user_a):
    response = admin_client.get("%s&username=*ael" % users_admin_link)
    assert_contains(response, user_a.username)


def test_search_by_username_end_excludes_users_with_different_username(
    admin_client, users_admin_link, user_b, user_c
):
    response = admin_client.get("%s&username=*ael" % users_admin_link)
    assert_not_contains(response, user_b.username)
    assert_not_contains(response, user_c.username)


def test_search_finds_users_by_username_content(
    admin_client, users_admin_link, user_a, user_b
):
    response = admin_client.get("%s&username=*yr*" % users_admin_link)
    assert_contains(response, user_a.username)
    assert_contains(response, user_b.username)


def test_search_by_username_content_excludes_users_with_different_username(
    admin_client, users_admin_link, user_c
):
    response = admin_client.get("%s&username=*yr*" % users_admin_link)
    assert_not_contains(response, user_c.username)


def test_search_finds_user_by_email(admin_client, users_admin_link, user_a):
    response = admin_client.get("%s&email=test123@test.org" % users_admin_link)
    assert_contains(response, user_a.email)


def test_search_excludes_users_with_different_email(
    admin_client, users_admin_link, user_b, user_c
):
    response = admin_client.get("%s&email=test123@test.org" % users_admin_link)
    assert_not_contains(response, user_b.email)
    assert_not_contains(response, user_c.email)


def test_search_finds_users_by_email_start(
    admin_client, users_admin_link, user_a, user_b
):
    response = admin_client.get("%s&email=test*" % users_admin_link)
    assert_contains(response, user_a.email)
    assert_contains(response, user_b.email)


def test_search_by_email_start_excludes_users_with_different_email(
    admin_client, users_admin_link, user_c
):
    response = admin_client.get("%s&email=test*" % users_admin_link)
    assert_not_contains(response, user_c.email)


def test_search_finds_user_by_email_end(admin_client, users_admin_link, user_a):
    response = admin_client.get("%s&email=*org" % users_admin_link)
    assert_contains(response, user_a.email)


def test_search_by_email_end_excludes_users_with_different_email(
    admin_client, users_admin_link, user_b, user_c
):
    response = admin_client.get("%s&email=*org" % users_admin_link)
    assert_not_contains(response, user_b.email)
    assert_not_contains(response, user_c.email)


def test_search_finds_users_by_email_content(
    admin_client, users_admin_link, user_b, user_c
):
    response = admin_client.get("%s&email=*@gmail*" % users_admin_link)
    assert_contains(response, user_b.email)
    assert_contains(response, user_c.email)


def test_search_by_email_content_excludes_users_with_different_email(
    admin_client, users_admin_link, user_a
):
    response = admin_client.get("%s&email=*@gmail*" % users_admin_link)
    assert_not_contains(response, user_a.email)


@pytest.fixture
def rank(db):
    return Rank.objects.create(name="Test Rank")


def test_search_finds_user_with_rank(admin_client, users_admin_link, rank):
    user = create_test_user("UserWithRank", "rank@example.org", rank=rank)
    response = admin_client.get("%s&rank=%s" % (users_admin_link, rank.pk))
    assert_contains(response, user.username)


def test_staff_users_search_excludes_user_without_rank(
    admin_client, users_admin_link, rank
):
    user = create_test_user("RegularUser", "regular@example.org")
    response = admin_client.get("%s&rank=%s" % (users_admin_link, rank.pk))
    assert_not_contains(response, user.username)


@pytest.fixture
def role(db):
    return Role.objects.create(name="Test Role")


def test_search_finds_user_with_role(admin_client, users_admin_link, role):
    user = create_test_user("UserWithRole", "role@example.org")
    user.roles.add(role)
    response = admin_client.get("%s&role=%s" % (users_admin_link, role.pk))
    assert_contains(response, user.username)


def test_staff_users_search_excludes_user_without_role(
    admin_client, users_admin_link, role
):
    user = create_test_user("RegularUser", "regular@example.org")
    response = admin_client.get("%s&role=%s" % (users_admin_link, role.pk))
    assert_not_contains(response, user.username)


def test_search_finds_inactive_user(admin_client, users_admin_link):
    user = create_test_user(
        "InactiveUser", "inactive@example.org", requires_activation=1
    )
    response = admin_client.get("%s&is_inactive=1" % users_admin_link)
    assert_contains(response, user.username)


def test_inactive_users_search_excludes_activated_users(admin_client, users_admin_link):
    user = create_test_user(
        "ActivatedUser", "activated@example.org", requires_activation=0
    )
    response = admin_client.get("%s&is_inactive=1" % users_admin_link)
    assert_not_contains(response, user.username)


def test_search_finds_disabled_user(admin_client, users_admin_link):
    user = create_test_user("DisabledUser", "disabled@example.org", is_active=False)
    response = admin_client.get("%s&is_disabled=1" % users_admin_link)
    assert_contains(response, user.username)


def test_disabled_users_search_excludes_active_users(admin_client, users_admin_link):
    user = create_test_user("ActiveUser", "active@example.org", is_active=True)
    response = admin_client.get("%s&is_disabled=1" % users_admin_link)
    assert_not_contains(response, user.username)


def test_search_finds_staff_user(admin_client, users_admin_link):
    user = create_test_user("StaffUser", "staff@example.org", is_staff=True)
    response = admin_client.get("%s&is_staff=1" % users_admin_link)
    assert_contains(response, user.username)


def test_staff_users_search_excludes_non_staff_users(admin_client, users_admin_link):
    user = create_test_user("RegularUser", "non_staff@example.org", is_staff=False)
    response = admin_client.get("%s&is_staff=1" % users_admin_link)
    assert_not_contains(response, user.username)


def test_search_finds_user_deleting_account(admin_client, users_admin_link):
    user = create_test_user(
        "DeletingUser", "deleting@example.org", is_deleting_account=True
    )
    response = admin_client.get("%s&is_deleting_account=1" % users_admin_link)
    assert_contains(response, user.username)


def test_staff_users_search_excludes_non_deleting_users(admin_client, users_admin_link):
    user = create_test_user(
        "RegularUser", "regular@example.org", is_deleting_account=False
    )
    response = admin_client.get("%s&is_deleting_account=1" % users_admin_link)
    assert_not_contains(response, user.username)



from typing import TYPE_CHECKING, Optional

from django.db.models import F

from ..categories.models import Category
from ..threads.models import Post, Thread
from .models import Notification

if TYPE_CHECKING:
    from ..users.models import User


def notify_user(
    user: "User",
    verb: str,
    actor: Optional["User"] = None,
    category: Optional[Category] = None,
    thread: Optional[Thread] = None,
    post: Optional[Post] = None,
) -> Notification:
    notification = Notification.objects.create(
        user=user,
        verb=verb,
        actor=actor,
        actor_name=actor.username if actor else None,
        category=category,
        thread=thread,
        thread_title=thread.title if thread else None,
        post=post,
    )

    user.unread_notifications = F("unread_notifications") + 1
    user.save(update_fields=["unread_notifications"])

    return notification



from datetime import timedelta

import pytest
from django.utils import timezone

from ...threads.models import ThreadParticipant
from ...users.bans import ban_user
from ..models import Notification
from ..tasks import notify_on_new_thread_reply


@pytest.fixture
def notify_watcher_mock(mocker):
    return mocker.patch("misago.notifications.tasks.notify_watcher_on_new_thread_reply")


def test_notify_on_new_thread_reply_does_nothing_for_unwatched_thread(
    notify_watcher_mock, user_reply
):
    notify_on_new_thread_reply(user_reply.id)
    notify_watcher_mock.assert_not_called()


def test_notify_on_new_thread_reply_skips_reply_author(
    watched_thread_factory, notify_watcher_mock, user, thread, user_reply
):
    watched_thread_factory(user, thread, send_emails=True)
    notify_on_new_thread_reply(user_reply.id)
    notify_watcher_mock.assert_not_called()


def test_notify_on_new_thread_reply_skips_watcher_with_deactivated_account(
    watched_thread_factory, notify_watcher_mock, inactive_user, thread, user_reply
):
    watched_thread_factory(inactive_user, thread, send_emails=True)
    notify_on_new_thread_reply(user_reply.id)
    notify_watcher_mock.assert_not_called()


def test_notify_on_new_thread_reply_skips_watcher_with_banned_account(
    watched_thread_factory, notify_watcher_mock, other_user, thread, user_reply
):
    ban_user(other_user)
    watched_thread_factory(other_user, thread, send_emails=True)

    notify_on_new_thread_reply(user_reply.id)
    notify_watcher_mock.assert_not_called()


def test_notify_on_new_thread_reply_handles_exceptions(
    mocker, watched_thread_factory, other_user, thread, user_reply
):
    notify_watcher_mock = mocker.patch(
        "misago.notifications.tasks.notify_watcher_on_new_thread_reply",
        side_effect=ValueError("Unknown"),
    )

    watched_thread_factory(other_user, thread, send_emails=True)

    notify_on_new_thread_reply(user_reply.id)
    notify_watcher_mock.assert_called_once()


def test_notify_on_new_thread_reply_notifies_user_about_thread_reply(
    watched_thread_factory, user, other_user, thread, user_reply, mailoutbox
):
    watched_thread_factory(other_user, thread, send_emails=False)
    notify_on_new_thread_reply(user_reply.id)

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 1

    Notification.objects.get(user=other_user, actor=user)
    assert len(mailoutbox) == 0


def test_notify_on_new_thread_reply_notifies_user_with_email_about_thread_reply(
    watched_thread_factory, user, other_user, thread, user_reply, mailoutbox
):
    watched_thread_factory(other_user, thread, send_emails=True)
    notify_on_new_thread_reply(user_reply.id)

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 1

    Notification.objects.get(user=other_user, actor=user)
    assert len(mailoutbox) == 1


def test_notify_on_new_thread_reply_checks_user_thread_permissions(
    mocker, watched_thread_factory, other_user, thread, user_reply, mailoutbox
):
    can_see_thread_mock = mocker.patch(
        "misago.notifications.threads.can_see_thread", return_value=False
    )

    watched_thread_factory(other_user, thread, send_emails=True)
    notify_on_new_thread_reply(user_reply.id)

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 0

    assert not Notification.objects.exists()
    assert len(mailoutbox) == 0

    can_see_thread_mock.assert_called_once()


def test_notify_on_new_thread_reply_checks_user_has_no_older_unread_posts(
    watched_thread_factory, other_user, thread, user_reply, mailoutbox
):
    watched_thread = watched_thread_factory(other_user, thread, send_emails=True)

    # Make thread's first post unread
    watched_thread.read_at = timezone.now() - timedelta(seconds=5)
    watched_thread.save()

    notify_on_new_thread_reply(user_reply.id)

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 0

    assert not Notification.objects.exists()
    assert len(mailoutbox) == 0


def test_notify_on_new_thread_reply_excludes_user_posts_from_unread_check(
    watched_thread_factory, user, other_user, thread, post, user_reply, mailoutbox
):
    watched_thread = watched_thread_factory(other_user, thread, send_emails=True)

    # Make thread's first post unread
    watched_thread.read_at = timezone.now() - timedelta(seconds=5)
    watched_thread.save()

    # Make thread's first post author the watcher
    post.poster = other_user
    post.save()

    notify_on_new_thread_reply(user_reply.id)

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 1

    Notification.objects.get(user=other_user, actor=user)
    assert len(mailoutbox) == 1


def test_notify_on_new_thread_reply_notifies_user_with_email_about_private_thread_reply(
    watched_thread_factory,
    other_user,
    private_thread,
    user,
    private_thread_user_reply,
    mailoutbox,
):
    ThreadParticipant.objects.create(thread=private_thread, user=other_user)

    watched_thread_factory(other_user, private_thread, send_emails=True)
    notify_on_new_thread_reply(private_thread_user_reply.id)

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 1

    Notification.objects.get(user=other_user, actor=user)
    assert len(mailoutbox) == 1


def test_notify_on_new_thread_reply_checks_if_user_is_private_thread_participant(
    watched_thread_factory,
    other_user,
    private_thread,
    private_thread_user_reply,
    mailoutbox,
):
    watched_thread_factory(other_user, private_thread, send_emails=True)
    notify_on_new_thread_reply(private_thread_user_reply.id)

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 0

    assert not Notification.objects.exists()
    assert len(mailoutbox) == 0



import re

from markdown.blockprocessors import HRProcessor


class BBCodeHRProcessor(HRProcessor):
    RE = r"^\[hr\]*"

    # Detect hr on any line of a block.
    SEARCH_RE = re.compile(RE, re.MULTILINE | re.IGNORECASE)



from django.contrib.admin import ModelAdmin
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext as _


class UserAdminModel(ModelAdmin):
    """
    The model should be used for interaction of third party django apps with
    Misago's `User`.

    Removes `new` and `delete` actions (use Misago admin for that).

    Registration call is placed in :mod:`misago.users.admin`.
    The tests are in :mod:`misago.users.tests.test_djangoadmin_user`.
    """

    list_display = (
        "username",
        "email",
        "rank",
        "is_staff",
        "is_superuser",
        "get_edit_from_misago_url",
    )
    search_fields = ("username", "email")
    list_filter = ("groups", "rank", "is_staff", "is_superuser")

    actions = None
    readonly_fields = (
        "username",
        "email",
        "joined_on",
        "last_login",
        "rank",
        "is_staff",
        "is_superuser",
        "get_edit_from_misago_url",
    )
    fieldsets = (
        (
            _("Misago user data"),
            {
                "fields": (
                    "username",
                    "email",
                    "joined_on",
                    "last_login",
                    "rank",
                    "is_staff",
                    "is_superuser",
                    "get_edit_from_misago_url",
                )
            },
        ),
        (_("Edit permissions and groups"), {"fields": ("groups", "user_permissions")}),
    )

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def get_edit_from_misago_url(self, user_instance):
        return format_html(
            '<a href="{link}" class="{cls}" target="blank">{text}</a>',
            link=reverse(
                viewname="misago:admin:users:edit", kwargs={"pk": user_instance.pk}
            ),
            cls="changelink",
            text=_("Edit"),
        )

    get_edit_from_misago_url.short_description = _(
        "Edit the user from Misago admin panel"
    )



from django.test import TestCase
from django.urls import reverse

from ...conf.test import override_dynamic_settings


class AuthViewsTests(TestCase):
    def test_auth_views_return_302(self):
        """auth views should always return redirect"""
        response = self.client.get(reverse("misago:login"))
        self.assertEqual(response.status_code, 302)

        response = self.client.post(reverse("misago:login"))
        self.assertEqual(response.status_code, 302)

        response = self.client.get(reverse("misago:logout"))
        self.assertEqual(response.status_code, 302)

        response = self.client.post(reverse("misago:logout"))
        self.assertEqual(response.status_code, 302)

    def test_login_view_redirect_to(self):
        """login view respects redirect_to POST"""
        # valid redirect
        response = self.client.post(
            reverse("misago:login"), data={"redirect_to": "/redirect/"}
        )

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response["location"], "/redirect/?ref=login")

        # invalid redirect (redirects to other site)
        response = self.client.post(
            reverse("misago:login"),
            data={"redirect_to": "http://somewhereelse.com/page.html"},
        )

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response["location"], "/")

        # invalid redirect (link name)
        response = self.client.post(
            reverse("misago:login"), data={"redirect_to": "misago:users"}
        )

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response["location"], "/")

        # invalid redirect (non url)
        response = self.client.post(
            reverse("misago:login"), data={"redirect_to": "canada goose not url!"}
        )

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response["location"], "/")

        # invalid redirect (unicode)
        response = self.client.post(
            reverse("misago:login"), data={"redirect_to": "łelcome!"}
        )

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response["location"], "/")

    def test_logout_view(self):
        """logout view logs user out on post"""
        response = self.client.post(
            "/api/auth/", data={"username": "nope", "password": "not-checked"}
        )

        self.assertContains(
            response, "Login or password is incorrect.", status_code=400
        )

        response = self.client.get("/api/auth/")
        self.assertEqual(response.status_code, 200)

        user_json = response.json()
        self.assertIsNone(user_json["id"])

        response = self.client.post(reverse("misago:logout"))
        self.assertEqual(response.status_code, 302)

        response = self.client.get("/api/auth/")
        self.assertEqual(response.status_code, 200)

        user_json = response.json()
        self.assertIsNone(user_json["id"])


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_provider="Lorem",
)
def test_login_view_returns_403_if_oauth_is_enabled(db, client):
    response = client.get(reverse("misago:login"))
    assert response.status_code == 403



from django.test import TestCase

from ...users.test import UserTestCase
from ..models import Agreement, UserAgreement
from ..utils import (
    get_parsed_agreement_text,
    get_required_user_agreement,
    save_user_agreement_acceptance,
)


class MockRequest:
    def __init__(self, user=None):
        self.user = user
        self.frontend_context = {}

    def get_host(self):
        return "testhost.com"


class GetParsedAgreementTextTests(TestCase):
    def test_agreement_no_text(self):
        agreement = Agreement.objects.create(
            type=Agreement.TYPE_PRIVACY, link="https://somewhre.com", is_active=True
        )

        result = get_parsed_agreement_text(MockRequest(), agreement)
        self.assertIsNone(result)

    def test_agreement_link_and_text(self):
        agreement = Agreement.objects.create(
            type=Agreement.TYPE_PRIVACY,
            link="https://somewhre.com",
            text="Lorem ipsum",
            is_active=True,
        )

        result = get_parsed_agreement_text(MockRequest(), agreement)
        self.assertEqual(result, "<p>Lorem ipsum</p>")

    def test_agreement_text(self):
        agreement = Agreement.objects.create(
            type=Agreement.TYPE_PRIVACY, text="Lorem ipsum", is_active=True
        )

        result = get_parsed_agreement_text(MockRequest(), agreement)
        self.assertEqual(result, "<p>Lorem ipsum</p>")


class GetRequiredUserAgreementTests(UserTestCase):
    def setUp(self):
        self.agreement = Agreement.objects.create(
            type=Agreement.TYPE_PRIVACY,
            link="https://somewhre.com",
            text="Lorem ipsum",
            is_active=True,
        )

        self.agreements = Agreement.objects.get_agreements_from_db()

    def test_anonymous_user(self):
        anonymous_user = self.get_anonymous_user()
        result = get_required_user_agreement(anonymous_user, self.agreements)
        self.assertIsNone(result)

    def test_authenticated_user_no_agreements(self):
        authenticated_user = self.get_authenticated_user()
        result = get_required_user_agreement(authenticated_user, {})
        self.assertIsNone(result)

    def test_authenticated_user(self):
        authenticated_user = self.get_authenticated_user()
        result = get_required_user_agreement(authenticated_user, self.agreements)
        self.assertEqual(result, self.agreement)

    def test_authenticated_user_with_agreement(self):
        authenticated_user = self.get_authenticated_user()
        authenticated_user.agreements.append(self.agreement.pk)

        result = get_required_user_agreement(authenticated_user, self.agreements)
        self.assertIsNone(result)

    def test_prioritize_terms_of_service(self):
        terms_of_service = Agreement.objects.create(
            type=Agreement.TYPE_TOS,
            link="https://somewhre.com",
            text="Lorem ipsum",
            is_active=True,
        )

        agreements = Agreement.objects.get_agreements_from_db()
        agreements_in_wrong_order = {
            Agreement.TYPE_PRIVACY: agreements[Agreement.TYPE_PRIVACY],
            Agreement.TYPE_TOS: agreements[Agreement.TYPE_TOS],
        }

        authenticated_user = self.get_authenticated_user()
        result = get_required_user_agreement(
            authenticated_user, agreements_in_wrong_order
        )
        self.assertEqual(result, terms_of_service)


class SaveUserAgreementAcceptance(UserTestCase):
    def test_no_commit(self):
        user = self.get_authenticated_user()

        agreement = Agreement.objects.create(
            type=Agreement.TYPE_PRIVACY, link="https://somewhre.com", text="Lorem ipsum"
        )

        save_user_agreement_acceptance(user, agreement)
        self.assertEqual(user.agreements, [agreement.id])

        user.refresh_from_db()
        self.assertEqual(user.agreements, [])

        UserAgreement.objects.get(user=user, agreement=agreement)
        self.assertEqual(UserAgreement.objects.count(), 1)

    def test_commit(self):
        user = self.get_authenticated_user()

        agreement = Agreement.objects.create(
            type=Agreement.TYPE_PRIVACY, link="https://somewhre.com", text="Lorem ipsum"
        )

        save_user_agreement_acceptance(user, agreement, commit=True)
        self.assertEqual(user.agreements, [agreement.id])

        user.refresh_from_db()
        self.assertEqual(user.agreements, [agreement.id])

        UserAgreement.objects.get(user=user, agreement=agreement)
        self.assertEqual(UserAgreement.objects.count(), 1)



from django.core import mail
from django.test import TestCase
from django.urls import reverse

from ...conf.test import override_dynamic_settings
from ..models import Ban
from ..test import create_test_user
from ..tokens import make_password_change_token


class GatewayTests(TestCase):
    def test_api_invalid_credentials(self):
        """login api returns 400 on invalid POST"""
        response = self.client.post(
            "/api/auth/", data={"username": "nope", "password": "nope"}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"code": "invalid_login", "detail": "Login or password is incorrect."},
        )

        response = self.client.get("/api/auth/")
        self.assertEqual(response.status_code, 200)

        user_json = response.json()
        self.assertIsNone(user_json["id"])

    def test_login(self):
        """api signs user in"""
        user = create_test_user("User", "user@example.com", "password")

        response = self.client.post(
            "/api/auth/", data={"username": "User", "password": "password"}
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.get("/api/auth/")
        self.assertEqual(response.status_code, 200)

        user_json = response.json()
        self.assertEqual(user_json["id"], user.id)
        self.assertEqual(user_json["username"], user.username)

    def test_login_whitespaces_password(self):
        """api signs user in with password left untouched"""
        user = create_test_user("User", "user@example.com", " password ")

        response = self.client.post(
            "/api/auth/", data={"username": "User", "password": "password"}
        )
        self.assertEqual(response.status_code, 400)

        response = self.client.post(
            "/api/auth/", data={"username": "User", "password": " password "}
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.get("/api/auth/")
        self.assertEqual(response.status_code, 200)

        user_json = response.json()
        self.assertEqual(user_json["id"], user.id)
        self.assertEqual(user_json["username"], user.username)

    def test_submit_empty(self):
        """login api errors for no body"""
        response = self.client.post("/api/auth/")
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"code": "empty_data", "detail": "Fill out both fields."}
        )

    def test_submit_invalid(self):
        """login api errors for invalid data"""
        response = self.client.post(
            "/api/auth/", "false", content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "non_field_errors": [
                    "Invalid data. Expected a dictionary, but got bool."
                ]
            },
        )

    def test_login_not_usable_password(self):
        """login api fails to sign user with not-usable password in"""
        create_test_user("User", "user@example.com")

        response = self.client.post(
            "/api/auth/", data={"username": "User", "password": "password"}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"code": "invalid_login", "detail": "Login or password is incorrect."},
        )

    def test_login_banned(self):
        """login api fails to sign banned user in"""
        create_test_user("User", "user@example.com", "password")

        ban = Ban.objects.create(
            check_type=Ban.USERNAME,
            banned_value="user",
            user_message="You are tragically banned.",
        )

        response = self.client.post(
            "/api/auth/", data={"username": "User", "password": "password"}
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(response_json["code"], "banned")
        self.assertEqual(response_json["detail"]["message"]["plain"], ban.user_message)
        self.assertEqual(
            response_json["detail"]["message"]["html"], "<p>%s</p>" % ban.user_message
        )

        response = self.client.get("/api/auth/")
        self.assertEqual(response.status_code, 200)

        user_json = response.json()
        self.assertIsNone(user_json["id"])

    def test_login_banned_staff(self):
        """login api signs banned staff member in"""
        user = create_test_user("User", "user@example.com", "password")
        user.is_staff = True
        user.save()

        Ban.objects.create(
            check_type=Ban.USERNAME,
            banned_value="user",
            user_message="You are tragically banned.",
        )

        response = self.client.post(
            "/api/auth/", data={"username": "User", "password": "password"}
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.get("/api/auth/")
        self.assertEqual(response.status_code, 200)

        user_json = response.json()
        self.assertEqual(user_json["id"], user.id)
        self.assertEqual(user_json["username"], user.username)

    def test_login_ban_registration_only(self):
        """login api ignores registration-only bans"""
        user = create_test_user("User", "user@example.com", "password")

        Ban.objects.create(
            check_type=Ban.USERNAME, banned_value="user", registration_only=True
        )

        response = self.client.post(
            "/api/auth/", data={"username": "User", "password": "password"}
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.get("/api/auth/")
        self.assertEqual(response.status_code, 200)

        user_json = response.json()
        self.assertEqual(user_json["id"], user.id)
        self.assertEqual(user_json["username"], user.username)

    def test_login_inactive_admin(self):
        """login api fails to sign admin-activated user in"""
        create_test_user("User", "user@example.com", "password", requires_activation=1)

        response = self.client.post(
            "/api/auth/", data={"username": "User", "password": "password"}
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(response_json["code"], "inactive_user")

        response = self.client.get("/api/auth/")
        self.assertEqual(response.status_code, 200)

        user_json = response.json()
        self.assertIsNone(user_json["id"])

    def test_login_inactive_user(self):
        """login api fails to sign user-activated user in"""
        create_test_user("User", "user@example.com", "password", requires_activation=2)

        response = self.client.post(
            "/api/auth/", data={"username": "User", "password": "password"}
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(response_json["code"], "inactive_admin")

        response = self.client.get("/api/auth/")
        self.assertEqual(response.status_code, 200)

        user_json = response.json()
        self.assertIsNone(user_json["id"])

    def test_login_disabled_user(self):
        """its impossible to sign in to disabled account"""
        user = create_test_user("User", "user@example.com", "password", is_active=False)
        user.is_staff = True
        user.save()

        response = self.client.post(
            "/api/auth/", data={"username": "User", "password": "password"}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"code": "invalid_login", "detail": "Login or password is incorrect."},
        )

        response = self.client.get("/api/auth/")
        self.assertEqual(response.status_code, 200)

        user_json = response.json()
        self.assertIsNone(user_json["id"])


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_provider="Lorem",
)
def test_login_api_returns_403_if_oauth_is_enabled(user, user_password, client):
    response = client.post(
        reverse("misago:api:auth"),
        {"username": user.username, "password": user_password},
    )
    assert response.status_code == 403


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_provider="Lorem",
)
def test_auth_api_returns_user_if_oauth_is_enabled(user_client):
    response = user_client.get(reverse("misago:api:auth"))
    assert response.status_code == 200


class UserCredentialsTests(TestCase):
    def test_edge_returns_response(self):
        """api edge has no showstoppers"""
        response = self.client.get("/api/auth/criteria/")
        self.assertEqual(response.status_code, 200)


class SendActivationApiTests(TestCase):
    def setUp(self):
        self.user = create_test_user("User", "user@example.com", "password")
        self.user.requires_activation = 1
        self.user.save()

        self.link = "/api/auth/send-activation/"

    def test_submit_valid(self):
        """request activation link api sends reset link mail"""
        response = self.client.post(self.link, data={"email": self.user.email})
        self.assertEqual(response.status_code, 200)

        self.assertIn("Activate User", mail.outbox[0].subject)

    def test_submit_banned(self):
        """request activation link api passes for banned users"""
        Ban.objects.create(
            check_type=Ban.USERNAME,
            banned_value=self.user.username,
            user_message="Nope!",
        )

        response = self.client.post(self.link, data={"email": self.user.email})
        self.assertEqual(response.status_code, 200)

        self.assertIn("Activate User", mail.outbox[0].subject)

    def test_submit_disabled(self):
        """request activation link api fails disabled users"""
        self.user.is_active = False
        self.user.save()

        response = self.client.post(self.link, data={"email": self.user.email})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"code": "not_found", "detail": "No user with this e-mail exists."},
        )

        self.assertTrue(not mail.outbox)

    def test_submit_empty(self):
        """request activation link api errors for no body"""
        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"code": "empty_email", "detail": "Enter e-mail address."}
        )

        self.assertTrue(not mail.outbox)

    def test_submit_invalid_data(self):
        """login api errors for invalid data"""
        response = self.client.post(self.link, "false", content_type="application/json")
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "non_field_errors": [
                    "Invalid data. Expected a dictionary, but got bool."
                ]
            },
        )

    def test_submit_invalid_email(self):
        """request activation link api errors for invalid e-mail"""
        response = self.client.post(self.link, data={"email": "fake@mail.com"})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"code": "not_found", "detail": "No user with this e-mail exists."},
        )

        self.assertTrue(not mail.outbox)

    def test_submit_active_user(self):
        """request activation link api errors for active user"""
        self.user.requires_activation = 0
        self.user.save()

        response = self.client.post(self.link, data={"email": self.user.email})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "code": "already_active",
                "detail": "User, your account is already active.",
            },
        )

    def test_submit_inactive_user(self):
        """request activation link api errors for admin-activated users"""
        self.user.requires_activation = 2
        self.user.save()

        response = self.client.post(self.link, data={"email": self.user.email})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "code": "inactive_admin",
                "detail": "User, only administrator may activate your account.",
            },
        )

        self.assertTrue(not mail.outbox)

        # but succeed for user-activated
        self.user.requires_activation = 1
        self.user.save()

        response = self.client.post(self.link, data={"email": self.user.email})
        self.assertEqual(response.status_code, 200)

        self.assertTrue(mail.outbox)


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_provider="Lorem",
)
def test_send_activation_api_returns_403_if_oauth_is_enabled(user, client):
    response = client.post(
        reverse("misago:api:send-activation"),
        {"email": user.email},
    )
    assert response.status_code == 403


class SendPasswordFormApiTests(TestCase):
    def setUp(self):
        self.user = create_test_user("User", "user@example.com", "password")

        self.link = "/api/auth/send-password-form/"

    def test_submit_valid(self):
        """request change password form link api sends reset link mail"""
        response = self.client.post(self.link, data={"email": self.user.email})
        self.assertEqual(response.status_code, 200)

        self.assertIn("Change User password", mail.outbox[0].subject)

    def test_submit_banned(self):
        """request change password form link api sends reset link mail"""
        Ban.objects.create(
            check_type=Ban.USERNAME,
            banned_value=self.user.username,
            user_message="Nope!",
        )

        response = self.client.post(self.link, data={"email": self.user.email})
        self.assertEqual(response.status_code, 200)

        self.assertIn("Change User password", mail.outbox[0].subject)

    def test_submit_disabled(self):
        """request change password form api fails disabled users"""
        self.user.is_active = False
        self.user.save()

        response = self.client.post(self.link, data={"email": self.user.email})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"code": "not_found", "detail": "No user with this e-mail exists."},
        )

        self.assertTrue(not mail.outbox)

    def test_submit_empty(self):
        """request change password form link api errors for no body"""
        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"code": "empty_email", "detail": "Enter e-mail address."}
        )

        self.assertTrue(not mail.outbox)

    def test_submit_invalid(self):
        """request change password form link api errors for invalid e-mail"""
        response = self.client.post(self.link, data={"email": "fake@mail.com"})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"code": "not_found", "detail": "No user with this e-mail exists."},
        )

        self.assertTrue(not mail.outbox)

    def test_submit_invalid_data(self):
        """login api errors for invalid data"""
        response = self.client.post(self.link, "false", content_type="application/json")
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "non_field_errors": [
                    "Invalid data. Expected a dictionary, but got bool."
                ]
            },
        )

    def test_submit_inactive_user(self):
        """request change password form link api errors for inactive users"""
        self.user.requires_activation = 1
        self.user.save()

        response = self.client.post(self.link, data={"email": self.user.email})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "code": "inactive_user",
                "detail": (
                    "You have to activate your account before you "
                    "will be able to request new password."
                ),
            },
        )

        self.user.requires_activation = 2
        self.user.save()

        response = self.client.post(self.link, data={"email": self.user.email})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "code": "inactive_admin",
                "detail": (
                    "Administrator has to activate your account before you "
                    "will be able to request new password."
                ),
            },
        )

        self.assertTrue(not mail.outbox)


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_provider="Lorem",
)
def test_send_password_reset_api_returns_403_if_oauth_is_enabled(user, client):
    response = client.post(
        reverse("misago:api:send-password-form"),
        {"email": user.email},
    )
    assert response.status_code == 403


class ChangePasswordApiTests(TestCase):
    def setUp(self):
        self.user = create_test_user("User", "user@example.com", "password")

        self.link = "/api/auth/change-password/%s/%s/"

    def test_submit_valid(self):
        """submit change password form api changes password"""
        response = self.client.post(
            self.link % (self.user.pk, make_password_change_token(self.user)),
            data={"password": "n3wp4ss!"},
        )
        self.assertEqual(response.status_code, 200)

        self.user.refresh_from_db()
        self.assertTrue(self.user.check_password("n3wp4ss!"))

    def test_submit_with_whitespaces(self):
        """submit change password form api changes password with whitespaces"""
        response = self.client.post(
            self.link % (self.user.pk, make_password_change_token(self.user)),
            data={"password": " n3wp4ss! "},
        )
        self.assertEqual(response.status_code, 200)

        self.user.refresh_from_db()
        self.assertTrue(self.user.check_password(" n3wp4ss! "))

    def test_submit_invalid_data(self):
        """login api errors for invalid data"""
        response = self.client.post(
            self.link % (self.user.pk, make_password_change_token(self.user)),
            "false",
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "non_field_errors": [
                    "Invalid data. Expected a dictionary, but got bool."
                ]
            },
        )

    def test_invalid_token_link(self):
        """api errors on invalid user id link"""
        response = self.client.post(self.link % (self.user.pk, "asda7ad89sa7d9s789as"))
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "Form link is invalid. Please request new one."}
        )

    def test_banned_user_link(self):
        """request errors because user is banned"""
        Ban.objects.create(
            check_type=Ban.USERNAME,
            banned_value=self.user.username,
            user_message="Nope!",
        )

        response = self.client.post(
            self.link % (self.user.pk, make_password_change_token(self.user))
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "Your link has expired. Please request new one."},
        )

    def test_inactive_user(self):
        """change password api errors for inactive users"""
        self.user.requires_activation = 1
        self.user.save()

        response = self.client.post(
            self.link % (self.user.pk, make_password_change_token(self.user))
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "Your link has expired. Please request new one."},
        )

        self.user.requires_activation = 2
        self.user.save()

        response = self.client.post(
            self.link % (self.user.pk, make_password_change_token(self.user))
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "Your link has expired. Please request new one."},
        )

    def test_disabled_user(self):
        """change password api errors for disabled users"""
        self.user.is_active = False
        self.user.save()

        response = self.client.post(
            self.link % (self.user.pk, make_password_change_token(self.user))
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "Form link is invalid. Please request new one."}
        )

    def test_submit_empty(self):
        """change password api errors for empty body"""
        response = self.client.post(
            self.link % (self.user.pk, make_password_change_token(self.user))
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "detail": (
                    "This password is too short. It must contain at least 7 characters."
                )
            },
        )


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_provider="Lorem",
)
def test_reset_password_api_returns_403_if_oauth_is_enabled(user, client):
    token = make_password_change_token(user)

    response = client.post(
        reverse("misago:api:change-forgotten-password", args=[user.pk, token]),
        {"password": "n33wP4SSW00ird!!"},
    )
    assert response.status_code == 403



from rest_framework import serializers

from . import PostingEndpoint, PostingMiddleware
from ... import moderation


class HideMiddleware(PostingMiddleware):
    def use_this_middleware(self):
        return self.mode == PostingEndpoint.START

    def get_serializer(self):
        return HideSerializer(data=self.request.data)

    def post_save(self, serializer):
        if self.thread.category.acl["can_hide_threads"]:
            if serializer.validated_data.get("hide"):
                moderation.hide_thread(self.request, self.thread)
                self.thread.update_all = True
                self.thread.save(update_fields=["is_hidden"])

                self.thread.category.synchronize()
                self.thread.category.update_all = True


class HideSerializer(serializers.Serializer):
    hide = serializers.BooleanField(required=False, default=False)



import pytest
from django.urls import reverse

from ...acl.models import Role
from ...acl.test import patch_user_acl
from ...conf import settings
from ...conf.test import override_dynamic_settings
from ..models import Attachment, AttachmentType
from ..test import post_thread


@pytest.fixture
def attachment_type(db):
    return AttachmentType.objects.order_by("id").first()


@pytest.fixture
def attachment(attachment_type, post, user):
    return Attachment.objects.create(
        secret="secret",
        filetype=attachment_type,
        post=post,
        uploader=user,
        uploader_name=user.username,
        uploader_slug=user.slug,
        filename="test.txt",
        file="test.txt",
        size=1000,
    )


@pytest.fixture
def image(post, user):
    return Attachment.objects.create(
        secret="secret",
        filetype=AttachmentType.objects.get(mimetypes="image/png"),
        post=post,
        uploader=user,
        uploader_name=user.username,
        uploader_slug=user.slug,
        filename="test.png",
        image="test.png",
        size=1000,
    )


@pytest.fixture
def image_with_thumbnail(post, user):
    return Attachment.objects.create(
        secret="secret",
        filetype=AttachmentType.objects.get(mimetypes="image/png"),
        post=post,
        uploader=user,
        uploader_name=user.username,
        uploader_slug=user.slug,
        filename="test.png",
        image="test.png",
        thumbnail="test-thumbnail.png",
        size=1000,
    )


@pytest.fixture
def other_users_attachment(attachment, other_user):
    attachment.uploader = other_user
    attachment.save()
    return attachment


@pytest.fixture
def orphaned_attachment(attachment):
    attachment.post = None
    attachment.save()
    return attachment


@pytest.fixture
def other_users_orphaned_attachment(other_users_attachment):
    other_users_attachment.post = None
    other_users_attachment.save()
    return other_users_attachment


def assert_403(response):
    assert response.status_code == 302
    assert response["location"].endswith(settings.MISAGO_ATTACHMENT_403_IMAGE)


def assert_404(response):
    assert response.status_code == 302
    assert response["location"].endswith(settings.MISAGO_ATTACHMENT_404_IMAGE)


def test_proxy_redirects_client_to_attachment_file(client, attachment):
    response = client.get(attachment.get_absolute_url())
    assert response.status_code == 302
    assert response["location"].endswith("test.txt")


def test_proxy_redirects_client_to_attachment_image(client, image):
    response = client.get(image.get_absolute_url())
    assert response.status_code == 302
    assert response["location"].endswith("test.png")


def test_proxy_redirects_client_to_attachment_thumbnail(client, image_with_thumbnail):
    response = client.get(image_with_thumbnail.get_thumbnail_url())
    assert response.status_code == 302
    assert response["location"].endswith("test-thumbnail.png")


def test_proxy_redirects_to_404_image_for_nonexistant_attachment(db, client):
    response = client.get(
        reverse("misago:attachment", kwargs={"pk": 1, "secret": "secret"})
    )
    assert_404(response)


def test_proxy_redirects_to_404_image_for_url_with_invalid_attachment_secret(
    client, attachment
):
    response = client.get(
        reverse("misago:attachment", kwargs={"pk": attachment.id, "secret": "invalid"})
    )
    assert_404(response)


@patch_user_acl({"can_download_other_users_attachments": False})
def test_proxy_redirects_to_403_image_for_user_without_permission_to_see_attachment(
    user_client, other_users_attachment
):
    response = user_client.get(other_users_attachment.get_absolute_url())
    assert_403(response)


def test_thumbnail_proxy_redirects_to_404_for_non_image_attachment(client, attachment):
    response = client.get(
        reverse(
            "misago:attachment-thumbnail",
            kwargs={"pk": attachment.pk, "secret": attachment.secret},
        )
    )
    assert_404(response)


def test_thumbnail_proxy_redirects_to_regular_image_for_image_without_thumbnail(
    client, image
):
    response = client.get(
        reverse(
            "misago:attachment-thumbnail",
            kwargs={"pk": image.pk, "secret": image.secret},
        )
    )
    assert response.status_code == 302
    assert response["location"].endswith("test.png")


def test_thumbnail_proxy_redirects_to_thumbnail_image(client, image_with_thumbnail):
    response = client.get(
        reverse(
            "misago:attachment-thumbnail",
            kwargs={
                "pk": image_with_thumbnail.pk,
                "secret": image_with_thumbnail.secret,
            },
        )
    )
    assert response.status_code == 302
    assert response["location"].endswith("test-thumbnail.png")


def test_proxy_blocks_user_from_their_orphaned_attachment(
    user_client, orphaned_attachment
):
    response = user_client.get(orphaned_attachment.get_absolute_url())
    assert_404(response)


def test_proxy_redirects_user_to_their_orphaned_attachment_if_link_has_shva_key(
    user_client, orphaned_attachment
):
    response = user_client.get("%s?shva=1" % orphaned_attachment.get_absolute_url())
    assert response.status_code == 302
    assert response["location"].endswith("test.txt")


def test_proxy_blocks_user_from_other_users_orphaned_attachment(
    user_client, other_users_orphaned_attachment
):
    response = user_client.get(other_users_orphaned_attachment.get_absolute_url())
    assert_404(response)


def test_proxy_blocks_user_from_other_users_orphaned_attachment_if_link_has_shva_key(
    user_client, other_users_orphaned_attachment
):
    response = user_client.get(
        "%s?shva=1" % other_users_orphaned_attachment.get_absolute_url()
    )
    assert_404(response)


def test_proxy_redirects_staff_to_other_users_orphaned_attachment(
    staff_client, orphaned_attachment
):
    response = staff_client.get("%s?shva=1" % orphaned_attachment.get_absolute_url())
    assert response.status_code == 302
    assert response["location"].endswith("test.txt")


def test_proxy_blocks_user_from_attachment_with_disabled_type(
    user_client, attachment, attachment_type
):
    attachment_type.status = AttachmentType.DISABLED
    attachment_type.save()

    response = user_client.get(attachment.get_absolute_url())
    assert_403(response)


@pytest.fixture
def role(db):
    return Role.objects.create(name="Test")


@pytest.fixture
def limited_attachment_type(attachment_type, role):
    attachment_type.limit_downloads_to.add(role)
    return attachment_type


def test_proxy_blocks_user_without_role_from_attachment_with_limited_type(
    user_client, attachment, limited_attachment_type
):
    response = user_client.get(attachment.get_absolute_url())
    assert_403(response)


def test_proxy_allows_user_with_role_to_download_attachment_with_limited_type(
    user, user_client, role, attachment, limited_attachment_type
):
    user.roles.add(role)
    response = user_client.get(attachment.get_absolute_url())
    assert response.status_code == 302
    assert response["location"].endswith("test.txt")


def test_proxy_allows_staff_user_without_role_to_download_attachment_with_limited_type(
    staff_client, role, attachment, limited_attachment_type
):
    response = staff_client.get(attachment.get_absolute_url())
    assert response.status_code == 302
    assert response["location"].endswith("test.txt")


@override_dynamic_settings(attachment_403_image="custom-403-image.png")
@patch_user_acl({"can_download_other_users_attachments": False})
def test_proxy_uses_custom_permission_denied_image_if_one_is_set(
    user_client, other_users_attachment
):
    response = user_client.get(other_users_attachment.get_absolute_url())
    assert response.status_code == 302
    assert response["location"].endswith("custom-403-image.png")


@override_dynamic_settings(attachment_404_image="custom-404-image.png")
def test_proxy_uses_custom_not_found_image_if_one_is_set(db, client):
    response = client.get(
        reverse("misago:attachment", kwargs={"pk": 1, "secret": "secret"})
    )
    assert response.status_code == 302
    assert response["location"].endswith("custom-404-image.png")



from ...admin.views import render
from ..metadata import plugins_metadata


def plugins_list(request):
    return render(
        request,
        "misago/admin/plugins/list.html",
        {"plugins": plugins_metadata.get_metadata().values()},
    )



from django.urls import reverse
from django.utils.crypto import get_random_string
from django.utils.translation import npgettext, pgettext
from rest_framework import serializers

from ..models import Poll

__all__ = [
    "PollSerializer",
    "EditPollSerializer",
    "NewPollSerializer",
    "PollChoiceSerializer",
]

MAX_POLL_OPTIONS = 16


class PollSerializer(serializers.ModelSerializer):
    acl = serializers.SerializerMethodField()
    choices = serializers.SerializerMethodField()

    api = serializers.SerializerMethodField()
    url = serializers.SerializerMethodField()

    class Meta:
        model = Poll
        fields = [
            "id",
            "poster_name",
            "posted_on",
            "length",
            "question",
            "allowed_choices",
            "allow_revotes",
            "votes",
            "is_public",
            "acl",
            "choices",
            "api",
            "url",
        ]

    def get_api(self, obj):
        return {"index": obj.get_api_url(), "votes": obj.get_votes_api_url()}

    def get_url(self, obj):
        return {"poster": self.get_poster_url(obj)}

    def get_poster_url(self, obj):
        if obj.poster_id:
            return reverse(
                "misago:user", kwargs={"slug": obj.poster_slug, "pk": obj.poster_id}
            )

    def get_acl(self, obj):
        try:
            return obj.acl
        except AttributeError:
            return None

    def get_choices(self, obj):
        return obj.choices


class EditPollSerializer(serializers.ModelSerializer):
    length = serializers.IntegerField(required=True, min_value=0, max_value=180)
    question = serializers.CharField(required=True, max_length=255)
    allowed_choices = serializers.IntegerField(required=True, min_value=1)
    choices = serializers.ListField(allow_empty=False, child=serializers.DictField())

    class Meta:
        model = Poll
        fields = ["length", "question", "allowed_choices", "allow_revotes", "choices"]

    def validate_choices(self, choices):
        clean_choices = list(map(self.clean_choice, choices))

        # generate hashes for added choices
        choices_map = {}
        for choice in self.instance.choices:
            choices_map[choice["hash"]] = choice

        final_choices = []
        for choice in clean_choices:
            if choice["hash"] in choices_map:
                choices_map[choice["hash"]].update({"label": choice["label"]})
                final_choices.append(choices_map[choice["hash"]])
            else:
                choice.update({"hash": get_random_string(12), "votes": 0})
                final_choices.append(choice)

        self.validate_choices_num(final_choices)

        return final_choices

    def clean_choice(self, choice):
        clean_choice = {
            "hash": choice.get("hash", get_random_string(12)),
            "label": choice.get("label", ""),
        }

        serializer = PollChoiceSerializer(data=clean_choice)
        if not serializer.is_valid():
            raise serializers.ValidationError(
                pgettext(
                    "edit poll serializer", "One or more poll choices are invalid."
                )
            )

        return serializer.data

    def validate_choices_num(self, choices):
        total_choices = len(choices)

        if total_choices < 2:
            raise serializers.ValidationError(
                pgettext(
                    "edit poll serializer",
                    "You need to add at least two choices to a poll.",
                )
            )

        if total_choices > MAX_POLL_OPTIONS:
            # pylint: disable=line-too-long
            message = npgettext(
                "edit poll serializer",
                "You can't add more than %(limit_value)s option to a single poll (added %(show_value)s).",
                "You can't add more than %(limit_value)s options to a single poll (added %(show_value)s).",
                MAX_POLL_OPTIONS,
            )
            raise serializers.ValidationError(
                message % {"limit_value": MAX_POLL_OPTIONS, "show_value": total_choices}
            )

    def validate(self, data):
        if data["allowed_choices"] > len(data["choices"]):
            raise serializers.ValidationError(
                pgettext(
                    "edit poll serializer",
                    "Number of allowed choices can't be greater than number of all choices.",
                )
            )
        return data

    def update(self, instance, validated_data):
        if instance.choices:
            self.update_choices(instance, validated_data["choices"])

        return super().update(instance, validated_data)

    def update_choices(self, instance, cleaned_choices):
        removed_hashes = []

        final_hashes = [c["hash"] for c in cleaned_choices]
        for choice in instance.choices:
            if choice["hash"] not in final_hashes:
                instance.votes -= choice["votes"]
                removed_hashes.append(choice["hash"])

        if removed_hashes:
            instance.pollvote_set.filter(choice_hash__in=removed_hashes).delete()


class NewPollSerializer(EditPollSerializer):
    class Meta:
        model = Poll
        fields = [
            "length",
            "question",
            "allowed_choices",
            "allow_revotes",
            "is_public",
            "choices",
        ]

    def validate_choices(self, choices):
        clean_choices = list(map(self.clean_choice, choices))

        self.validate_choices_num(clean_choices)

        for choice in clean_choices:
            choice.update({"hash": get_random_string(12), "votes": 0})

        return clean_choices


class PollChoiceSerializer(serializers.Serializer):
    hash = serializers.CharField(required=True, min_length=12, max_length=12)
    label = serializers.CharField(required=True, max_length=255)



from ..threads import ThreadNotifications, get_watched_threads


def test_get_watched_threads_returns_empty_dict_if_threads_are_not_watched(
    user, thread, other_thread
):
    watched_threads = get_watched_threads(user, [thread, other_thread])
    assert watched_threads == {}


def test_get_watched_threads_returns_watched_threads_for_user(
    user, other_user, thread, other_thread, watched_thread_factory
):
    watched_thread_factory(other_user, thread, send_emails=True)
    watched_thread_factory(user, thread, send_emails=False)
    watched_thread_factory(user, other_thread, send_emails=True)

    watched_threads = get_watched_threads(user, [thread, other_thread])
    assert watched_threads == {
        thread.id: ThreadNotifications.SITE_ONLY,
        other_thread.id: ThreadNotifications.SITE_AND_EMAIL,
    }


def test_get_watched_threads_excludes_unspecified_threads(
    user, thread, other_thread, watched_thread_factory
):
    watched_thread_factory(user, other_thread, send_emails=True)

    watched_threads = get_watched_threads(user, [thread])
    assert watched_threads == {}


def test_get_watched_threads_returns_first_watched_thread_for_user(
    user, thread, watched_thread_factory
):
    watched_thread_factory(user, thread, send_emails=False)
    watched_thread_factory(user, thread, send_emails=True)

    watched_threads = get_watched_threads(user, [thread])
    assert watched_threads == {thread.id: ThreadNotifications.SITE_ONLY}



from ...core.shortcuts import paginate, pagination_dict
from ..online.utils import make_users_status_aware
from ..serializers import UserCardSerializer


class RankUsers:
    def __init__(self, request, rank, page=0):
        queryset = rank.user_set.select_related(
            "rank", "ban_cache", "online_tracker"
        ).order_by("slug")

        if not request.user.is_staff:
            queryset = queryset.filter(is_active=True)

        list_page = paginate(
            queryset,
            page,
            request.settings.users_per_page,
            request.settings.users_per_page_orphans,
        )
        make_users_status_aware(request, list_page.object_list)

        self.users = list_page.object_list
        self.paginator = pagination_dict(list_page)

    def get_frontend_context(self):
        context = {"results": UserCardSerializer(self.users, many=True).data}
        context.update(self.paginator)
        return context

    def get_template_context(self):
        return {"users": self.users, "paginator": self.paginator}



from django import forms
from django.contrib import messages
from django.utils.translation import pgettext, pgettext_lazy

from ....admin.forms import YesNoSwitch
from .base import SettingsForm

OAUTH2_OPTIONAL_FIELDS = (
    "oauth2_token_extra_headers",
    "oauth2_user_extra_headers",
    "oauth2_send_welcome_email",
    "oauth2_json_avatar_path",
)


class OAuth2SettingsForm(SettingsForm):
    settings = [
        "enable_oauth2_client",
        "oauth2_provider",
        "oauth2_client_id",
        "oauth2_client_secret",
        "oauth2_scopes",
        "oauth2_login_url",
        "oauth2_token_url",
        "oauth2_token_extra_headers",
        "oauth2_json_token_path",
        "oauth2_user_url",
        "oauth2_user_method",
        "oauth2_user_token_location",
        "oauth2_user_token_name",
        "oauth2_user_extra_headers",
        "oauth2_send_welcome_email",
        "oauth2_json_id_path",
        "oauth2_json_name_path",
        "oauth2_json_email_path",
        "oauth2_json_avatar_path",
    ]

    enable_oauth2_client = YesNoSwitch(
        label=pgettext_lazy("admin oauth2 settings form", "Enable OAuth2 client"),
        help_text=pgettext_lazy(
            "admin oauth2 settings form",
            "Enabling OAuth2 will make login option redirect users to the OAuth provider configured below. It will also disable option to register on forum, change username, email or password, as those features will be delegated to the 3rd party site.",
        ),
    )
    oauth2_provider = forms.CharField(
        label=pgettext_lazy("admin oauth2 settings form", "Provider name"),
        help_text=pgettext_lazy(
            "admin oauth2 settings form",
            "Name of the OAuth 2 provider to be displayed by interface.",
        ),
        max_length=255,
        required=False,
    )
    oauth2_client_id = forms.CharField(
        label=pgettext_lazy("admin oauth2 settings form", "Client ID"),
        max_length=200,
        required=False,
    )
    oauth2_client_secret = forms.CharField(
        label=pgettext_lazy("admin oauth2 settings form", "Client Secret"),
        max_length=200,
        required=False,
    )
    oauth2_scopes = forms.CharField(
        label=pgettext_lazy("admin oauth2 settings form", "Scopes"),
        help_text=pgettext_lazy(
            "admin oauth2 settings form",
            "List of scopes to request from provider, separated with spaces.",
        ),
        max_length=500,
        required=False,
    )

    oauth2_login_url = forms.URLField(
        label=pgettext_lazy("admin oauth2 settings form", "Login form URL"),
        help_text=pgettext_lazy(
            "admin oauth2 settings form",
            "Address of the login form on provider's server that users will be redirected to.",
        ),
        max_length=500,
        required=False,
    )

    oauth2_token_url = forms.URLField(
        label=pgettext_lazy("admin oauth2 settings form", "Access token retrieval URL"),
        help_text=pgettext_lazy(
            "admin oauth2 settings form",
            "URL that will be called after user completes the login process and authorization code is sent back to your site. This URL is expected to take this code and return the access token that will be next used to retrieve user data.",
        ),
        max_length=500,
        required=False,
    )
    oauth2_token_extra_headers = forms.CharField(
        label=pgettext_lazy(
            "admin oauth2 settings form", "Extra HTTP headers in token request"
        ),
        help_text=pgettext_lazy(
            "admin oauth2 settings form",
            'List of extra headers to include in a HTTP request made to retrieve the access token. Example header is "Header-name: value". Specify each header on separate line.',
        ),
        widget=forms.Textarea(attrs={"rows": 4}),
        max_length=500,
        required=False,
    )
    oauth2_json_token_path = forms.CharField(
        label=pgettext_lazy("admin oauth2 settings form", "JSON path to access token"),
        help_text=pgettext_lazy(
            "admin oauth2 settings form",
            'Name of key containing the access token in JSON returned by the provider. If token is nested, use period (".") for path, eg: "result.token" will retrieve the token from "token" key nested in "result".',
        ),
        max_length=500,
        required=False,
    )

    oauth2_user_url = forms.URLField(
        label=pgettext_lazy("admin oauth2 settings form", "User data URL"),
        max_length=500,
        required=False,
    )
    oauth2_user_method = forms.ChoiceField(
        label=pgettext_lazy("admin oauth2 settings form", "Request method"),
        choices=[
            ("POST", "POST"),
            ("GET", "GET"),
        ],
        widget=forms.RadioSelect(),
    )
    oauth2_user_token_location = forms.ChoiceField(
        label=pgettext_lazy("admin oauth2 settings form", "Access token location"),
        choices=[
            (
                "QUERY",
                pgettext_lazy(
                    "admin oauth2 token location choice",
                    "Query string",
                ),
            ),
            (
                "HEADER",
                pgettext_lazy(
                    "admin oauth2 token location choice",
                    "HTTP header",
                ),
            ),
            (
                "HEADER_BEARER",
                pgettext_lazy(
                    "admin oauth2 token location choice",
                    "HTTP header (Bearer)",
                ),
            ),
        ],
        widget=forms.RadioSelect(),
    )
    oauth2_user_token_name = forms.CharField(
        label=pgettext_lazy("admin oauth2 settings form", "Access token name"),
        max_length=200,
        required=False,
    )
    oauth2_user_extra_headers = forms.CharField(
        label=pgettext_lazy(
            "admin oauth2 settings form", "Extra HTTP headers in user request"
        ),
        help_text=pgettext_lazy(
            "admin oauth2 settings form",
            'List of extra headers to include in a HTTP request made to retrieve the user profile. Example header is "Header-name: value". Specify each header on separate line.',
        ),
        widget=forms.Textarea(attrs={"rows": 4}),
        max_length=500,
        required=False,
    )

    oauth2_send_welcome_email = YesNoSwitch(
        label=pgettext_lazy(
            "admin oauth2 settings form",
            "Send a welcoming e-mail to users on their first sign-ons",
        ),
        required=False,
    )

    oauth2_json_id_path = forms.CharField(
        label=pgettext_lazy("admin oauth2 settings form", "User ID path"),
        max_length=200,
        required=False,
    )
    oauth2_json_name_path = forms.CharField(
        label=pgettext_lazy("admin oauth2 settings form", "User name path"),
        max_length=200,
        required=False,
    )
    oauth2_json_email_path = forms.CharField(
        label=pgettext_lazy("admin oauth2 settings form", "User e-mail path"),
        max_length=200,
        required=False,
    )
    oauth2_json_avatar_path = forms.CharField(
        label=pgettext_lazy("admin oauth2 settings form", "User avatar URL path"),
        help_text=pgettext_lazy(
            "admin oauth2 settings form",
            "Optional, leave empty to don't download avatar from provider.",
        ),
        max_length=200,
        required=False,
    )

    def clean_oauth2_scopes(self):
        # Remove duplicates and extra spaces, keep order of scopes
        clean_scopes = []
        for scope in self.cleaned_data["oauth2_scopes"].split():
            scope = scope.strip()
            if scope and scope not in clean_scopes:
                clean_scopes.append(scope)

        return " ".join(clean_scopes) or None

    def clean_oauth2_token_extra_headers(self):
        return clean_headers(self.cleaned_data["oauth2_token_extra_headers"])

    def clean_oauth2_user_extra_headers(self):
        return clean_headers(self.cleaned_data["oauth2_user_extra_headers"])

    def clean(self):
        data = super().clean()

        if not data.get("enable_oauth2_client"):
            return data

        required_data = [data[key] for key in data if key not in OAUTH2_OPTIONAL_FIELDS]

        if not all(required_data):
            data["enable_oauth2_client"] = False

            messages.error(
                self.request,
                pgettext(
                    "admin oauth2 settings form",
                    "You need to complete the configuration before you will be able to enable OAuth 2 on your site.",
                ),
            )

        return data


def clean_headers(headers_value):
    clean_headers = {}
    for header in headers_value.splitlines():
        header = header.strip()
        if not header:
            continue
        if ":" not in header:
            raise forms.ValidationError(
                pgettext(
                    "admin oauth2 settings form",
                    '"%(header)s" is not a valid header. It\'s missing a colon (":").',
                )
                % {"header": header},
            )

        name, value = [part.strip() for part in header.split(":", 1)]

        if not name:
            raise forms.ValidationError(
                pgettext(
                    "admin oauth2 settings form",
                    '"%(header)s" is not a valid header. It\'s missing a header name before the colon (":").',
                )
                % {"header": header},
            )

        if name in clean_headers:
            raise forms.ValidationError(
                pgettext(
                    "admin oauth2 settings form",
                    '"%(header)s" header is entered more than once.',
                )
                % {"header": name},
            )

        if not value:
            raise forms.ValidationError(
                pgettext(
                    "admin oauth2 settings form",
                    '"%(header)s" is not a valid header. It\'s missing a header value after the colon (":").',
                )
                % {"header": header},
            )

        clean_headers[name] = value

    return "\n".join([f"{key}: {value}" for key, value in clean_headers.items()])



from django import forms

from ...cache import clear_settings_cache


class SettingsForm(forms.Form):
    settings = []

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request")
        super().__init__(*args, **kwargs)

    def save(self, settings):
        self.save_settings(settings)
        self.clear_cache()

    def save_settings(self, settings):
        for setting in self.settings:
            setting_obj = settings[setting]
            new_value = self.cleaned_data.get(setting)
            if setting_obj.python_type == "image":
                if new_value and new_value != self.initial.get(setting):
                    self.save_image(setting_obj, new_value)
                elif self.cleaned_data.get("%s_delete" % setting):
                    self.delete_image(setting_obj)
            else:
                self.save_setting(setting_obj, new_value)

    def delete_image(self, setting):
        if setting.image:
            setting.image.delete()

    def save_image(self, setting, value):
        if setting.image:
            setting.image.delete(save=False)
        setting.value = value
        setting.save()

    def save_setting(self, setting, value):
        setting.value = value
        setting.save()

    def clear_cache(self):
        clear_settings_cache()



import time

from django.core.management.base import BaseCommand

from ....core.management.progressbar import show_progress
from ....core.pgutils import chunk_queryset
from ...checksums import update_post_checksum
from ...models import Post


class Command(BaseCommand):
    help = "Updates posts checksums"

    def handle(self, *args, **options):
        posts_to_update = Post.objects.filter(is_event=False).count()

        if not posts_to_update:
            self.stdout.write("\n\nNo posts were found")
        else:
            self.update_posts_checksums(posts_to_update)

    def update_posts_checksums(self, posts_to_update):
        self.stdout.write("Updating %s posts checksums...\n" % posts_to_update)

        updated_count = 0
        show_progress(self, updated_count, posts_to_update)
        start_time = time.time()

        queryset = Post.objects.filter(is_event=False)
        for post in chunk_queryset(queryset):
            update_post_checksum(post)
            post.save(update_fields=["checksum"])

            updated_count += 1
            show_progress(self, updated_count, posts_to_update, start_time)

        self.stdout.write("\n\nUpdated %s posts checksums" % updated_count)



from ..htmlparser import parse_html_string, print_html_string


def test_parser_handles_simple_html():
    root_node = parse_html_string("<p>Hello World!</p>")
    assert print_html_string(root_node) == "<p>Hello World!</p>"


def test_parser_handles_html_with_brs():
    root_node = parse_html_string("<p>Hello<br />World!</p>")
    assert print_html_string(root_node) == "<p>Hello<br />World!</p>"


def test_parser_handles_html_with_hrs():
    root_node = parse_html_string("<p>Hello</p><hr /><p>World!</p>")
    assert print_html_string(root_node) == "<p>Hello</p><hr /><p>World!</p>"


def test_parser_escapes_html_in_text_nodes():
    root_node = parse_html_string("<span>Hello &lt;br&gt; World!</span>")
    assert print_html_string(root_node) == "<span>Hello &lt;br&gt; World!</span>"


def test_parser_escapes_quotes_in_text_nodes():
    root_node = parse_html_string('<span>Hello "World"!</span>')
    assert print_html_string(root_node) == "<span>Hello &quot;World&quot;!</span>"


def test_parser_handles_attributes():
    root_node = parse_html_string('<a href="/hello-world/">Hello World!</a>')
    assert print_html_string(root_node) == '<a href="/hello-world/">Hello World!</a>'


def test_parser_escapes_html_in_attributes_names():
    root_node = parse_html_string('<span data-a<tt>r="<br>">Hello!</span>')
    assert print_html_string(root_node) == (
        "<span data-a&lt;tt>r=&quot;<br />&quot;&gt;Hello!</span>"
    )


def test_parser_escapes_quotes_in_attributes_names():
    root_node = parse_html_string('<span "data-attr"="br">Hello!</span>')
    assert print_html_string(root_node) == (
        '<span &quot;data-attr&quot;="br">Hello!</span>'
    )


def test_parser_escapes_html_in_attributes_values():
    root_node = parse_html_string('<span data-attr="<br>">Hello!</span>')
    assert print_html_string(root_node) == (
        '<span data-attr="&lt;br&gt;">Hello!</span>'
    )


def test_parser_handles_escaped_attribute_values():
    root_node = parse_html_string('<span data-attr="&lt;br&gt;">Hello!</span>')
    assert print_html_string(root_node) == (
        '<span data-attr="&lt;br&gt;">Hello!</span>'
    )


def test_parser_escapes_quotes_in_attributes_values():
    root_node = parse_html_string('<span data-attr="\'">Hello!</span>')
    assert print_html_string(root_node) == ('<span data-attr="&#x27;">Hello!</span>')


def test_parser_handles_bool_attributes():
    root_node = parse_html_string("<button disabled>Hello World!</button>")
    assert print_html_string(root_node) == "<button disabled>Hello World!</button>"



import pytest

from ..models import Theme


@pytest.fixture
def default_theme(db):
    return Theme.objects.get(is_default=True)


@pytest.fixture
def theme(db):
    return Theme.objects.create(name="Custom theme")


@pytest.fixture
def other_theme(db):
    return Theme.objects.create(name="Other theme")


@pytest.fixture
def active_theme(theme):
    Theme.objects.update(is_active=False)
    theme.is_active = True
    theme.save()
    return theme



# Generated by Django 3.2.15 on 2023-04-12 20:46

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import misago.notifications.models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("misago_threads", "0012_set_dj_partial_indexes"),
        ("misago_categories", "0009_auto_20221101_2111"),
    ]

    operations = [
        migrations.CreateModel(
            name="WatchedThread",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                ("send_emails", models.BooleanField(default=True)),
                (
                    "secret",
                    models.CharField(
                        default=misago.notifications.models.get_watched_thread_secret,
                        max_length=32,
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("read_at", models.DateTimeField(default=django.utils.timezone.now)),
                (
                    "category",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_categories.category",
                    ),
                ),
                (
                    "thread",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_threads.thread",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Notification",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                ("verb", models.CharField(max_length=32)),
                ("is_read", models.BooleanField(default=False)),
                ("actor_name", models.CharField(blank=True, max_length=255, null=True)),
                (
                    "thread_title",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                (
                    "actor",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "category",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_categories.category",
                    ),
                ),
                (
                    "post",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_threads.post",
                    ),
                ),
                (
                    "thread",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_threads.thread",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



"""
Service for tracking namechanges
"""
from datetime import timedelta

from django.utils import timezone


def get_username_options(settings, user, user_acl):
    changes_left = get_left_namechanges(user, user_acl)
    next_on = get_next_available_namechange(user, user_acl, changes_left)

    return {
        "changes_left": changes_left,
        "next_on": next_on,
        "length_min": settings.username_length_min,
        "length_max": settings.username_length_max,
    }


def get_left_namechanges(user, user_acl):
    name_changes_allowed = user_acl["name_changes_allowed"]
    if not name_changes_allowed:
        return 0

    valid_changes = get_valid_changes_queryset(user, user_acl)
    used_changes = valid_changes.count()
    if name_changes_allowed <= used_changes:
        return 0
    return name_changes_allowed - used_changes


def get_next_available_namechange(user, user_acl, changes_left):
    name_changes_expire = user_acl["name_changes_expire"]
    if changes_left or not name_changes_expire:
        return None

    valid_changes = get_valid_changes_queryset(user, user_acl)
    name_last_changed_on = valid_changes.latest().changed_on
    return name_last_changed_on + timedelta(days=name_changes_expire)


def get_valid_changes_queryset(user, user_acl):
    name_changes_expire = user_acl["name_changes_expire"]
    queryset = user.namechanges.filter(changed_by=user)
    if user_acl["name_changes_expire"]:
        cutoff = timezone.now() - timedelta(days=name_changes_expire)
        return queryset.filter(changed_on__gte=cutoff)
    return queryset



from django.db import models

from ..avatars import store


class AvatarGallery(models.Model):
    gallery = models.CharField(max_length=255)
    image = models.ImageField(max_length=255, upload_to=store.upload_to)

    class Meta:
        ordering = ["gallery", "pk"]

    @property
    def url(self):
        return self.image.url



from django.conf import settings
from django.db import models
from django.utils.crypto import get_random_string
from django.utils import timezone
from django.urls import reverse


WATCHED_THREAD_SECRET = 32


def get_watched_thread_secret() -> str:
    return get_random_string(WATCHED_THREAD_SECRET)


class Notification(models.Model):
    id = models.BigAutoField(primary_key=True)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)

    verb = models.CharField(max_length=32)
    is_read = models.BooleanField(default=False)

    actor = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )
    actor_name = models.CharField(max_length=255, blank=True, null=True)
    category = models.ForeignKey(
        "misago_categories.Category", blank=True, null=True, on_delete=models.CASCADE
    )
    thread = models.ForeignKey(
        "misago_threads.Thread", blank=True, null=True, on_delete=models.CASCADE
    )
    thread_title = models.CharField(max_length=255, blank=True, null=True)
    post = models.ForeignKey(
        "misago_threads.Post", blank=True, null=True, on_delete=models.CASCADE
    )

    created_at = models.DateTimeField(auto_now_add=True, db_index=True)

    def get_absolute_url(self) -> str:
        return reverse("misago:notification", kwargs={"notification_id": self.id})

    class Meta:
        indexes = [
            models.Index(fields=["-id", "user"]),
            models.Index(fields=["-id", "user", "is_read"]),
            models.Index(
                name="misago_noti_user_unread",
                fields=["user", "is_read"],
                condition=models.Q(is_read=False),
            ),
            models.Index(fields=["user", "post", "is_read"]),
        ]


class WatchedThread(models.Model):
    id = models.BigAutoField(primary_key=True)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    category = models.ForeignKey("misago_categories.Category", on_delete=models.CASCADE)
    thread = models.ForeignKey("misago_threads.Thread", on_delete=models.CASCADE)

    send_emails = models.BooleanField(default=True)
    secret = models.CharField(
        max_length=WATCHED_THREAD_SECRET, default=get_watched_thread_secret
    )

    created_at = models.DateTimeField(auto_now_add=True)
    read_at = models.DateTimeField(default=timezone.now)

    class Meta:
        indexes = [
            models.Index(fields=["user", "-thread"]),
        ]

    def get_disable_emails_url(self):
        return reverse(
            "misago:notifications-disable-email",
            kwargs={
                "watched_thread_id": self.id,
                "secret": self.secret,
            },
        )



from datetime import timedelta

import pytest
from django.utils import timezone

from ..audittrail import create_audit_trail, create_user_audit_trail
from ..models import AuditTrail
from ..signals import remove_old_ips

USER_IP = "13.41.51.41"


class RequestMock:
    user_ip = USER_IP

    def __init__(self, user):
        self.user = user


@pytest.fixture
def request_mock(user):
    return RequestMock(user)


def test_audit_trail_creation_raises_value_error_if_target_is_not_model_instance(
    request_mock, anonymous_user
):
    with pytest.raises(ValueError):
        create_audit_trail(request_mock, anonymous_user)


def test_audit_trail_is_not_created_for_anonymous_users(anonymous_user, user):
    request_mock = RequestMock(anonymous_user)
    create_audit_trail(request_mock, user)
    assert not AuditTrail.objects.exists()


def test_audit_trail_is_created(request_mock, other_user):
    assert create_audit_trail(request_mock, other_user)
    assert AuditTrail.objects.exists()


def test_audit_trail_is_created_with_request_data(request_mock, user, other_user):
    audit_trail = create_audit_trail(request_mock, other_user)

    assert audit_trail.user == user
    assert audit_trail.ip_address == USER_IP


def test_audit_trail_is_created_with_generic_relation_to_target(
    request_mock, user, other_user
):
    audit_trail = create_audit_trail(request_mock, other_user)
    assert audit_trail.content_object == other_user


def test_audit_trail_is_deleted_together_with_user(request_mock, user, other_user):
    audit_trail = create_audit_trail(request_mock, other_user)
    user.delete(anonymous_username="Deleted")
    with pytest.raises(AuditTrail.DoesNotExist):
        audit_trail.refresh_from_db()


def test_audit_trail_is_kept_after_its_target_is_deleted(request_mock, other_user):
    audit_trail = create_audit_trail(request_mock, other_user)
    other_user.delete(anonymous_username="Deleted")
    audit_trail.refresh_from_db()


def test_deleting_audit_trail_leaves_user(request_mock, user, other_user):
    audit_trail = create_audit_trail(request_mock, other_user)
    audit_trail.delete()
    user.refresh_from_db()


def test_deleting_audit_trail_leaves_target(request_mock, user, other_user):
    audit_trail = create_audit_trail(request_mock, other_user)
    audit_trail.delete()
    other_user.refresh_from_db()


def test_audit_trail_can_be_created_without_request(user, other_user):
    assert create_user_audit_trail(user, USER_IP, other_user)
    assert AuditTrail.objects.exists()


def test_audit_trail_creation_without_request_raises_value_error_if_target_is_not_model(
    user, anonymous_user
):
    with pytest.raises(ValueError):
        create_user_audit_trail(user, USER_IP, anonymous_user)


def test_audit_trail_without_request_is_created_with_explicit_data(user, other_user):
    audit_trail = create_user_audit_trail(user, USER_IP, other_user)

    assert audit_trail.user == user
    assert audit_trail.ip_address == USER_IP


def test_audit_trail_without_request_is_created_with_generic_relation_to_target(
    user, other_user
):
    audit_trail = create_user_audit_trail(user, USER_IP, other_user)
    assert audit_trail.content_object == other_user


def test_recent_audit_trail_is_not_deleted_on_signal(user, other_user):
    create_user_audit_trail(user, USER_IP, other_user)
    remove_old_ips.send(None, ip_storage_time=1)
    assert user.audittrail_set.exists()


def test_old_audit_trail_is_deleted_on_signal(user, other_user):
    audit_trail = create_user_audit_trail(user, USER_IP, other_user)
    audit_trail.created_on = timezone.now() - timedelta(days=6)
    audit_trail.save()

    remove_old_ips.send(None, ip_storage_time=5)
    assert not user.audittrail_set.exists()



from django.urls import reverse


class Node:
    def __init__(self, *, name=None, description=None, icon=None, link=None):
        self.parent = None
        self.name = name
        self.description = description
        self.icon = icon
        self.link = link
        self._children = []
        self._children_dict = {}
        self._resolved_namespace = None

    @property
    def namespace(self):
        if self._resolved_namespace:
            return self._resolved_namespace

        bits = self.link.split(":")
        self._resolved_namespace = ":".join(bits[:-1])
        return self._resolved_namespace

    def children(self):
        return self._children

    def children_as_dicts(self):
        childrens = []
        for children in self._children:
            childrens.append(
                {
                    "name": children.name,
                    "description": children.description,
                    "icon": children.icon,
                    "link": reverse(children.link),
                    "namespace": children.namespace,
                }
            )
        return childrens

    def add_node(self, node, after=None, before=None):
        if after:
            return self.add_node_after(node, after)
        if before:
            return self.add_node_before(node, before)
        node.parent = self
        self._children.append(node)
        self._children_dict[node.link] = node
        return True

    def add_node_after(self, node, after):
        success = False
        new_children_list = []

        for children in self._children:
            new_children_list.append(children)
            if children.link == after:
                new_children_list.append(node)
                success = True

        if success:
            node.parent = self
            self._children_dict[node.link] = node
            self._children = new_children_list
        return success

    def add_node_before(self, node, before):
        success = False
        new_children_list = []

        for children in self._children:
            if children.link == before:
                new_children_list.append(node)
                success = True
            new_children_list.append(children)

        if success:
            node.parent = self
            self._children_dict[node.link] = node
            self._children = new_children_list
        return success

    def child(self, namespace):
        try:
            return self._children_dict[namespace]
        except KeyError:
            raise ValueError(
                "Node %s is not a child of node %s" % (namespace, self.name)
            )

    def is_root(self):
        return False


class AdminSite:
    def __init__(self):
        self.nodes_record = []
        self.nodes_dict = {}

    def build_nodes_dict(self):
        nodes_dict = {"misago:admin": Node(link="misago:admin:index")}

        iterations = 0
        while self.nodes_record:
            iterations += 1
            if iterations > 512:
                message = (
                    "Misago Admin hierarchy is invalid or too complex to resolve. "
                    "Nodes left: %s"
                )
                raise ValueError(message % self.nodes_record)

            for index, node in enumerate(self.nodes_record):
                if node["parent"] in nodes_dict:
                    node_obj = Node(
                        name=node["name"],
                        description=node["description"],
                        icon=node["icon"],
                        link=node["link"],
                    )

                    parent = nodes_dict[node["parent"]]
                    if node["after"]:
                        node_added = parent.add_node(node_obj, after=node["after"])
                    elif node["before"]:
                        node_added = parent.add_node(node_obj, before=node["before"])
                    else:
                        node_added = parent.add_node(node_obj)

                    if node_added:
                        namespace = node.get("namespace") or node_obj.namespace

                        if namespace not in nodes_dict:
                            nodes_dict[namespace] = node_obj

                        del self.nodes_record[index]
                        break

        return nodes_dict

    def add_node(
        self,
        *,
        name=None,
        description=None,
        icon=None,
        parent=None,
        after=None,
        before=None,
        namespace=None,
        link="index",
    ):
        if self.nodes_dict:
            raise RuntimeError(
                "Misago admin site has already been initialized. "
                "You can't add new nodes to it."
            )

        if after and before:
            raise ValueError("after and before arguments are exclusive")

        self.nodes_record.append(
            {
                "name": name,
                "description": description,
                "icon": icon,
                "parent": join_namespace(parent),
                "after": join_namespace(parent, after) if after else None,
                "before": join_namespace(parent, before) if before else None,
                "namespace": join_namespace(parent, namespace),
                "link": join_namespace(parent, namespace, link),
            }
        )

    def visible_branches(self, request):
        if not self.nodes_dict:
            self.nodes_dict = self.build_nodes_dict()

        branches = []

        try:
            namespace = request.resolver_match.namespace
        except AttributeError:
            namespace = "misago:admin"

        if namespace in self.nodes_dict:
            node = self.nodes_dict[namespace]
            while node:
                children = node.children_as_dicts()
                if children:
                    branches.append(children)
                node = node.parent

        try:
            namespaces = request.resolver_match.namespaces
        except AttributeError:
            namespaces = ["misago", "admin"]

        branches.reverse()
        for depth, branch in enumerate(branches):
            depth_namespace = namespaces[2 : 3 + depth]
            for node in branch:
                node_namespace = node["namespace"].split(":")[2 : 3 + depth]
                if request.resolver_match:
                    node["is_active"] = depth_namespace == node_namespace
                else:
                    node["is_active"] = False

        return branches


def join_namespace(*args):
    parts = list(filter(None, args))
    parts.insert(0, "misago:admin")
    return ":".join(parts)


site = AdminSite()



from django.core import mail
from django.urls import reverse

from ...conf.test import override_dynamic_settings
from ..test import AuthenticatedUserTestCase, create_test_user


class UserChangeEmailTests(AuthenticatedUserTestCase):
    """tests for user change email RPC (/api/users/1/change-email/)"""

    def setUp(self):
        super().setUp()
        self.link = "/api/users/%s/change-email/" % self.user.pk

    def test_unsupported_methods(self):
        """api isn't supporting GET"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 405)

    def test_empty_input(self):
        """api errors correctly for empty input"""
        response = self.client.post(self.link, data={})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "new_email": ["This field is required."],
                "password": ["This field is required."],
            },
        )

    def test_invalid_password(self):
        """api errors correctly for invalid password"""
        response = self.client.post(
            self.link, data={"new_email": "new@email.com", "password": "Lor3mIpsum"}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"password": ["Entered password is invalid."]}
        )

    def test_invalid_input(self):
        """api errors correctly for invalid input"""
        response = self.client.post(
            self.link, data={"new_email": "", "password": self.USER_PASSWORD}
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"new_email": ["This field may not be blank."]}
        )

        response = self.client.post(
            self.link, data={"new_email": "newmail", "password": self.USER_PASSWORD}
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"new_email": ["Enter a valid e-mail address."]}
        )

    def test_email_taken(self):
        """api validates email usage"""
        taken_email = "otheruser@example.com"
        create_test_user("Other_User", taken_email)

        response = self.client.post(
            self.link, data={"new_email": taken_email, "password": self.USER_PASSWORD}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"new_email": ["This e-mail address is not available."]},
        )

    @override_dynamic_settings(forum_address="http://test.com/")
    def test_change_email(self):
        """api allows users to change their e-mail addresses"""
        new_email = "new@email.com"

        response = self.client.post(
            self.link, data={"new_email": new_email, "password": self.USER_PASSWORD}
        )
        self.assertEqual(response.status_code, 200)

        self.assertIn("Confirm e-mail change", mail.outbox[0].subject)
        for line in [l.strip() for l in mail.outbox[0].body.splitlines()]:
            if line.startswith("http://"):
                token = line.rstrip("/").split("/")[-1]
                break
        else:
            self.fail("E-mail sent didn't contain confirmation url")

        response = self.client.get(
            reverse("misago:options-confirm-email-change", kwargs={"token": token})
        )

        self.assertEqual(response.status_code, 200)

        self.reload_user()
        self.assertEqual(self.user.email, new_email)

    @override_dynamic_settings(forum_address="http://test.com/")
    def test_change_email_user_password_whitespace(self):
        """api supports users with whitespace around their passwords"""
        user_password = " old password "
        new_email = "new@email.com"

        self.user.set_password(user_password)
        self.user.save()

        self.login_user(self.user)

        response = self.client.post(
            self.link, data={"new_email": new_email, "password": user_password}
        )
        self.assertEqual(response.status_code, 200)

        self.assertIn("Confirm e-mail change", mail.outbox[0].subject)
        for line in [l.strip() for l in mail.outbox[0].body.splitlines()]:
            if line.startswith("http://"):
                token = line.rstrip("/").split("/")[-1]
                break
        else:
            self.fail("E-mail sent didn't contain confirmation url")

        response = self.client.get(
            reverse("misago:options-confirm-email-change", kwargs={"token": token})
        )

        self.assertEqual(response.status_code, 200)

        self.reload_user()
        self.assertEqual(self.user.email, new_email)

    @override_dynamic_settings(
        enable_oauth2_client=True,
        oauth2_provider="Lorem",
    )
    def test_change_email_api_returns_403_if_oauth_is_enabled(self):
        new_email = "new@email.com"

        self.login_user(self.user)

        response = self.client.post(
            self.link, data={"new_email": new_email, "password": self.USER_PASSWORD}
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(len(mail.outbox), 0)

    @override_dynamic_settings(forum_address="http://test.com/")
    def test_confirm_change_email_view_returns_403_if_oauth_is_enabled(self):
        new_email = "new@email.com"

        self.login_user(self.user)

        response = self.client.post(
            self.link, data={"new_email": new_email, "password": self.USER_PASSWORD}
        )
        self.assertEqual(response.status_code, 200)

        self.assertIn("Confirm e-mail change", mail.outbox[0].subject)
        for line in [l.strip() for l in mail.outbox[0].body.splitlines()]:
            if line.startswith("http://"):
                token = line.rstrip("/").split("/")[-1]
                break
        else:
            self.fail("E-mail sent didn't contain confirmation url")

        with override_dynamic_settings(
            enable_oauth2_client=True, oauth2_provider="Lorem"
        ):
            response = self.client.get(
                reverse("misago:options-confirm-email-change", kwargs={"token": token})
            )

            self.assertEqual(response.status_code, 403)



from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.urls import reverse

from ...acl.test import patch_user_acl
from ...categories.models import Category
from ...conf.test import override_dynamic_settings
from ...users.test import AuthenticatedUserTestCase, create_test_user
from ..models import ThreadParticipant
from ..test import other_user_cant_use_private_threads

User = get_user_model()


class StartPrivateThreadTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.private_threads()
        self.api_link = reverse("misago:api:private-thread-list")

        self.other_user = create_test_user("Other_User", "otheruser@example.com")

    def test_cant_start_thread_as_guest(self):
        """user has to be authenticated to be able to post private thread"""
        self.logout_user()

        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)

    @patch_user_acl({"can_use_private_threads": False})
    def test_cant_use_private_threads(self):
        """has no permission to use private threads"""
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "You can't use private threads."})

    @patch_user_acl({"can_start_private_threads": False})
    def test_cant_start_private_thread(self):
        """permission to start private thread is validated"""
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't start private threads."}
        )

    def test_empty_data(self):
        """no data sent handling has no showstoppers"""
        response = self.client.post(self.api_link, data={})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "to": ["This field is required."],
                "title": ["You have to enter a thread title."],
                "post": ["You have to enter a message."],
            },
        )

    def test_title_is_validated(self):
        """title is validated"""
        response = self.client.post(
            self.api_link,
            data={
                "to": [self.other_user.username],
                "title": "------",
                "post": "Lorem ipsum dolor met, sit amet elit!",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"title": ["Thread title should contain alpha-numeric characters."]},
        )

    def test_post_is_validated(self):
        """post is validated"""
        response = self.client.post(
            self.api_link,
            data={
                "to": [self.other_user.username],
                "title": "Lorem ipsum dolor met",
                "post": "a",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "post": [
                    "Posted message should be at least 5 characters long (it has 1)."
                ]
            },
        )

    def test_cant_invite_self(self):
        """api validates that you cant invite yourself to private thread"""
        response = self.client.post(
            self.api_link,
            data={
                "to": [self.user.username],
                "title": "Lorem ipsum dolor met",
                "post": "Lorem ipsum dolor.",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "to": [
                    "You can't include yourself on the list "
                    "of users to invite to new thread."
                ]
            },
        )

    def test_cant_invite_nonexisting(self):
        """api validates that you cant invite nonexisting user to thread"""
        response = self.client.post(
            self.api_link,
            data={
                "to": ["Ab", "Cd"],
                "title": "Lorem ipsum dolor met",
                "post": "Lorem ipsum dolor.",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"to": ["One or more users could not be found: ab, cd"]}
        )

    def test_cant_invite_too_many(self):
        """api validates that you cant invite too many users to thread"""
        response = self.client.post(
            self.api_link,
            data={
                "to": ["Username%s" % i for i in range(50)],
                "title": "Lorem ipsum dolor met",
                "post": "Lorem ipsum dolor.",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "to": [
                    "You can't add more than 3 users to private thread "
                    "(you've added 50)."
                ]
            },
        )

    @patch_user_acl(other_user_cant_use_private_threads)
    def test_cant_invite_no_permission(self):
        """api validates invited user permission to private thread"""
        response = self.client.post(
            self.api_link,
            data={
                "to": [self.other_user.username],
                "title": "Lorem ipsum dolor met",
                "post": "Lorem ipsum dolor.",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"to": ["Other_User can't participate in private threads."]},
        )

    def test_cant_invite_blocking(self):
        """api validates that you cant invite blocking user to thread"""
        self.other_user.blocks.add(self.user)

        response = self.client.post(
            self.api_link,
            data={
                "to": [self.other_user.username],
                "title": "Lorem ipsum dolor met",
                "post": "Lorem ipsum dolor.",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"to": ["Other_User is blocking you."]})

    @patch_user_acl({"can_add_everyone_to_private_threads": 1})
    def test_cant_invite_blocking_override(self):
        """api validates that you cant invite blocking user to thread"""
        self.other_user.blocks.add(self.user)

        response = self.client.post(
            self.api_link,
            data={
                "to": [self.other_user.username],
                "title": "-----",
                "post": "Lorem ipsum dolor.",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"title": ["Thread title should contain alpha-numeric characters."]},
        )

    def test_cant_invite_followers_only(self):
        """api validates that you cant invite followers-only user to thread"""
        user_constant = User.LIMIT_INVITES_TO_FOLLOWED
        self.other_user.limits_private_thread_invites_to = user_constant
        self.other_user.save()

        response = self.client.post(
            self.api_link,
            data={
                "to": [self.other_user.username],
                "title": "Lorem ipsum dolor met",
                "post": "Lorem ipsum dolor.",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "to": [
                    "Other_User limits invitations to private threads to followed users."
                ]
            },
        )

        # allow us to bypass following check
        with patch_user_acl({"can_add_everyone_to_private_threads": 1}):
            response = self.client.post(
                self.api_link,
                data={
                    "to": [self.other_user.username],
                    "title": "-----",
                    "post": "Lorem ipsum dolor.",
                },
            )

            self.assertEqual(response.status_code, 400)
            self.assertEqual(
                response.json(),
                {"title": ["Thread title should contain alpha-numeric characters."]},
            )

        # make user follow us
        self.other_user.follows.add(self.user)

        response = self.client.post(
            self.api_link,
            data={
                "to": [self.other_user.username],
                "title": "-----",
                "post": "Lorem ipsum dolor.",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"title": ["Thread title should contain alpha-numeric characters."]},
        )

    def test_cant_invite_anyone(self):
        """api validates that you cant invite nobody user to thread"""
        user_constant = User.LIMIT_INVITES_TO_NOBODY
        self.other_user.limits_private_thread_invites_to = user_constant
        self.other_user.save()

        response = self.client.post(
            self.api_link,
            data={
                "to": [self.other_user.username],
                "title": "Lorem ipsum dolor met",
                "post": "Lorem ipsum dolor.",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"to": ["Other_User has disabled invitations to private threads."]},
        )

        # allow us to bypass user preference check
        with patch_user_acl({"can_add_everyone_to_private_threads": 1}):
            response = self.client.post(
                self.api_link,
                data={
                    "to": [self.other_user.username],
                    "title": "-----",
                    "post": "Lorem ipsum dolor.",
                },
            )

            self.assertEqual(response.status_code, 400)
            self.assertEqual(
                response.json(),
                {"title": ["Thread title should contain alpha-numeric characters."]},
            )

    @override_dynamic_settings(forum_address="http://test.com/")
    @patch("misago.threads.participants.notify_on_new_private_thread")
    def test_can_start_thread(self, notify_on_new_private_thread_mock):
        """endpoint creates new thread"""
        response = self.client.post(
            self.api_link,
            data={
                "to": [self.other_user.username],
                "title": "Hello, I am test thread!",
                "post": "Lorem ipsum dolor met!",
            },
        )
        self.assertEqual(response.status_code, 200)

        thread = self.user.thread_set.all()[:1][0]

        response_json = response.json()
        self.assertEqual(response_json["url"], thread.get_absolute_url())

        response = self.client.get(thread.get_absolute_url())
        self.assertContains(response, self.category.name)
        self.assertContains(response, thread.title)
        self.assertContains(response, "<p>Lorem ipsum dolor met!</p>")

        # don't count private threads
        self.reload_user()
        self.assertEqual(self.user.threads, 0)
        self.assertEqual(self.user.posts, 0)

        self.assertEqual(thread.category_id, self.category.pk)
        self.assertEqual(thread.title, "Hello, I am test thread!")
        self.assertEqual(thread.starter_id, self.user.id)
        self.assertEqual(thread.starter_name, self.user.username)
        self.assertEqual(thread.starter_slug, self.user.slug)
        self.assertEqual(thread.last_poster_id, self.user.id)
        self.assertEqual(thread.last_poster_name, self.user.username)
        self.assertEqual(thread.last_poster_slug, self.user.slug)

        post = self.user.post_set.all()[:1][0]
        self.assertEqual(post.category_id, self.category.pk)
        self.assertEqual(post.original, "Lorem ipsum dolor met!")
        self.assertEqual(post.poster_id, self.user.id)
        self.assertEqual(post.poster_name, self.user.username)

        self.assertEqual(self.user.audittrail_set.count(), 1)

        # thread has two participants
        self.assertEqual(thread.participants.count(), 2)

        # we are thread owner
        ThreadParticipant.objects.get(thread=thread, user=self.user, is_owner=True)

        # other user was added to thread
        ThreadParticipant.objects.get(
            thread=thread, user=self.other_user, is_owner=False
        )

        # other user has sync_unread_private_threads flag
        self.other_user.refresh_from_db()
        self.assertTrue(self.other_user.sync_unread_private_threads)

        # notification about new private thread was triggered
        notify_on_new_private_thread_mock.delay.assert_called_once_with(
            self.user.id,
            thread.id,
            [self.other_user.id],
        )

    @patch("misago.threads.participants.notify_on_new_private_thread")
    def test_post_unicode(self, notify_on_new_private_thread_mock):
        """unicode characters can be posted"""
        response = self.client.post(
            self.api_link,
            data={
                "to": [self.other_user.username],
                "title": "Brzęczyżczykiewicz",
                "post": "Chrzążczyżewoszyce, powiat Łękółody.",
            },
        )
        self.assertEqual(response.status_code, 200)



from django.urls import reverse

from ...test import assert_contains


def test_notifications_view_is_accessible_by_users(user_client):
    response = user_client.get(reverse("misago:notifications"))
    assert_contains(response, "Notifications")
    assert_contains(response, "enable JavaScript")


def test_notifications_view_shows_permission_denied_page_to_guests(db, client):
    response = client.get(reverse("misago:notifications"))
    assert_contains(response, "You must be signed in", status_code=403)



from django.apps import AppConfig
from social_core.backends.facebook import FacebookOAuth2
from social_core.backends.github import GithubOAuth2
from social_core.backends.google import GoogleOAuth2
from social_core.backends.twitter import TwitterOAuth

from .providers import providers


class MisagoSocialAuthConfig(AppConfig):
    name = "misago.socialauth"
    label = "misago_socialauth"
    verbose_name = "Misago Social Auth"

    def ready(self):
        # Register default providers
        from .admin.forms import FacebookForm, GitHubForm, GoogleForm, TwitterForm

        providers.add(
            provider="facebook",
            name="Facebook",
            auth_backend=FacebookOAuth2,
            settings={"scope": ["email"]},
            admin_form=FacebookForm,
            admin_template="misago/admin/socialauth/form.html",
        )
        providers.add(
            provider="github",
            name="GitHub",
            auth_backend=GithubOAuth2,
            settings={"scope": ["read:user", "user:email"]},
            admin_form=GitHubForm,
            admin_template="misago/admin/socialauth/form.html",
        )
        providers.add(
            provider="google-oauth2",
            name="Google",
            auth_backend=GoogleOAuth2,
            admin_form=GoogleForm,
            admin_template="misago/admin/socialauth/form.html",
        )
        providers.add(
            provider="twitter",
            name="X",
            auth_backend=TwitterOAuth,
            admin_form=TwitterForm,
            admin_template="misago/admin/socialauth/form.html",
        )



import random

from django.utils import timezone

from ..threads.checksums import update_post_checksum
from ..threads.models import Post
from .englishcorpus import EnglishCorpus
from .users import get_fake_username

PLACEKITTEN_URL = "https://dummyimage.com/g/%s/%s"

corpus = EnglishCorpus()


def get_fake_post(fake, thread, poster=None):
    original, parsed = get_fake_post_content(fake)
    posted_on = timezone.now()

    post = Post.objects.create(
        category=thread.category,
        thread=thread,
        poster=poster,
        poster_name=poster.username if poster else get_fake_username(fake),
        original=original,
        parsed=parsed,
        posted_on=posted_on,
        updated_on=posted_on,
    )
    update_post_checksum(post)
    post.save(update_fields=["checksum"])

    return post


def get_fake_unapproved_post(fake, thread, poster=None):
    post = get_fake_post(fake, thread, poster)
    post.is_unapproved = True
    post.save(update_fields=["is_unapproved"])
    return post


def get_fake_hidden_post(fake, thread, poster=None, hidden_by=None):
    post = get_fake_post(fake, thread, poster)
    post.is_hidden = True

    if hidden_by:
        post.hidden_by = hidden_by
        post.hidden_by_name = hidden_by.username
        post.hidden_by_slug = hidden_by.slug
    else:
        post.hidden_by_name = fake.first_name()
        post.hidden_by_slug = post.hidden_by_name.lower()

    post.save(
        update_fields=["is_unapproved", "hidden_by", "hidden_by_name", "hidden_by_slug"]
    )

    return post


def get_fake_post_content(fake):
    raw = []
    parsed = []

    if random.randint(0, 100) > 90:
        paragraphs_to_make = random.randint(1, 20)
    else:
        paragraphs_to_make = random.randint(1, 5)

    for _ in range(paragraphs_to_make):
        if random.randint(0, 100) > 95:
            cat_width = random.randint(1, 16) * random.choice([100, 90, 80])
            cat_height = random.randint(1, 12) * random.choice([100, 90, 80])

            cat_url = PLACEKITTEN_URL % (cat_width, cat_height)

            raw.append("!(%s)" % cat_url)
            parsed.append('<p><img src="%s" alt=""/></p>' % cat_url)
        else:
            if random.randint(0, 100) > 95:
                sentences_to_make = random.randint(1, 20)
            else:
                sentences_to_make = random.randint(1, 7)
            raw.append(" ".join(corpus.random_sentences(sentences_to_make)))
            parsed.append("<p>%s</p>" % raw[-1])

    return "\n\n".join(raw), "\n".join(parsed)



from django import forms
from django.utils.translation import pgettext_lazy

from .base import OAuthProviderForm


class FacebookForm(OAuthProviderForm):
    key = forms.CharField(
        label=pgettext_lazy("admin social auth facebook form", "App ID"),
        required=False,
    )
    secret = forms.CharField(
        label=pgettext_lazy("admin social auth facebook form", "App Secret"),
        required=False,
    )



import re
from typing import Union

from django.contrib.auth import get_user_model

from ..users.utils import slugify_username
from .htmlparser import (
    ElementNode,
    RootNode,
    TextNode,
)

EXCLUDE_ELEMENTS = ("pre", "code", "a")
USERNAME_RE = re.compile(r"@[0-9a-z_]+", re.IGNORECASE)
MENTIONS_LIMIT = 32


def add_mentions(result, root_node):
    if "@" not in result["parsed_text"]:
        return

    mentions = set()  # usernames slugs
    nodes = []

    find_mentions(root_node, mentions, nodes)

    if not mentions or len(mentions) > MENTIONS_LIMIT:
        return  # No need to run mentions logic

    users_data = get_users_data(mentions)
    if not users_data:
        return  # Mentioned users don't exist

    for node in nodes:
        add_mentions_to_node(node, users_data)

    result["mentions"] = [user[0] for user in users_data.values()]


def find_mentions(
    node: Union[ElementNode, RootNode],
    mentions: set,
    nodes: set,
):
    if isinstance(node, ElementNode) and node.tag in EXCLUDE_ELEMENTS:
        return

    tracked_node = False
    for child in node.children:
        if isinstance(child, TextNode):
            results = find_mentions_in_str(child.text)
            if results:
                mentions.update(results)
                if not tracked_node:
                    tracked_node = True
                    nodes.append(node)
        else:
            find_mentions(child, mentions, nodes)


def find_mentions_in_str(text: str):
    matches = USERNAME_RE.findall(text)
    if not matches:
        return None

    return set([slugify_username(match[1:]) for match in matches])


def get_users_data(mentions):
    User = get_user_model()
    users_data = {}

    queryset = User.objects.filter(slug__in=mentions).values_list(
        "id", "username", "slug"
    )

    for user_id, username, slug in queryset:
        users_data[slug] = (user_id, username)

    return users_data


def add_mentions_to_node(node, users_data):
    new_children = []

    for child in node.children:
        if isinstance(child, TextNode):
            new_children += add_mentions_to_text(child.text, users_data)
        else:
            new_children.append(child)

    node.children = new_children


def add_mentions_to_text(text: str, users_data):
    nodes = []

    while True:
        match = USERNAME_RE.search(text)
        if not match:
            if text:
                nodes.append(TextNode(text=text))
            return nodes

        start, end = match.span()
        user_slug = text[start + 1 : end].lower().replace("_", "-")

        # Append text between 0 and start to nodes
        if start > 0:
            nodes.append(TextNode(text=text[:start]))

        # Append match string to nodes and keep scanning
        if user_slug not in users_data:
            nodes.append(TextNode(text=text[:end]))
            text = text[end:]
            continue

        user_id, username = users_data[user_slug]
        nodes.append(
            ElementNode(
                tag="a",
                attrs={
                    "href": f"/u/{user_slug}/{user_id}/",
                    "data-quote": f"@{username}",
                },
                children=[TextNode(text=f"@{username}")],
            )
        )

        text = text[end:]



from django.core.exceptions import PermissionDenied
from django.contrib.auth import get_user_model
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.utils.translation import pgettext

from ...core.exceptions import Banned
from ..bans import get_user_ban
from ..decorators import deny_authenticated, deny_banned_ips
from ..tokens import is_activation_token_valid

User = get_user_model()


def activation_view(f):
    @deny_authenticated
    @deny_banned_ips
    def decorator(request, *args, **kwargs):
        if request.settings.enable_oauth2_client:
            raise PermissionDenied(
                pgettext(
                    "activate account",
                    "Please use %(provider)s to activate your account.",
                )
                % {"provider": request.settings.oauth2_provider}
            )

        return f(request, *args, **kwargs)

    return decorator


@activation_view
def request_activation(request):
    request.frontend_context.update(
        {"SEND_ACTIVATION_API": reverse("misago:api:send-activation")}
    )
    return render(request, "misago/activation/request.html")


class ActivationStopped(Exception):
    pass


class ActivationError(Exception):
    pass


@activation_view
def activate_by_token(request, pk, token):
    inactive_user = get_object_or_404(User, pk=pk, is_active=True)

    try:
        if not inactive_user.requires_activation:
            message = pgettext(
                "activate account with token",
                "%(user)s, your account is already active.",
            )
            raise ActivationStopped(message % {"user": inactive_user.username})

        if not is_activation_token_valid(inactive_user, token):
            message = pgettext(
                "activate account with token",
                "%(user)s, your activation link is invalid. Try again or request new activation link.",
            )
            raise ActivationError(message % {"user": inactive_user.username})

        ban = get_user_ban(inactive_user, request.cache_versions)
        if ban:
            raise Banned(ban)
    except ActivationStopped as e:
        return render(request, "misago/activation/stopped.html", {"message": e.args[0]})
    except ActivationError as e:
        return render(
            request, "misago/activation/error.html", {"message": e.args[0]}, status=400
        )

    inactive_user.requires_activation = User.ACTIVATION_NONE
    inactive_user.save(update_fields=["requires_activation"])

    message = pgettext(
        "activate account with token", "%(user)s, your account has been activated!"
    )

    return render(
        request,
        "misago/activation/done.html",
        {"message": message % {"user": inactive_user.username}},
    )



from typing import Iterable

from django.http import HttpRequest

from ..categories.models import Category
from ..threads.models import Post, Thread
from ..threads.permissions import exclude_invisible_posts, exclude_invisible_threads
from .cutoffdate import get_cutoff_date


def get_categories_new_posts(
    request: HttpRequest,
    categories: Iterable[Category],
) -> dict[int, bool]:
    """Returns a dict with category ID as a key and bool if it has new posts."""
    if not categories:
        return {}

    categories_new_posts = {category.id: False for category in categories}

    if request.user.is_anonymous:
        return categories_new_posts

    threads = Thread.objects.filter(category__in=categories)
    threads = exclude_invisible_threads(request.user_acl, categories, threads)
    queryset = (
        Post.objects.filter(
            category__in=categories,
            thread__in=threads,
            posted_on__gt=get_cutoff_date(request.settings, request.user),
        )
        .values_list("category", flat=True)
        .distinct()
    )

    queryset = queryset.exclude(id__in=request.user.postread_set.values("post"))
    queryset = exclude_invisible_posts(request.user_acl, categories, queryset)

    for category_id in queryset:
        categories_new_posts[category_id] = True

    return categories_new_posts



from ....notifications.tasks import notify_on_new_thread_reply
from ....notifications.threads import (
    watch_replied_thread,
    watch_started_thread,
)
from . import PostingEndpoint, PostingMiddleware


class NotificationsMiddleware(PostingMiddleware):
    def use_this_middleware(self):
        return self.mode != PostingEndpoint.EDIT

    def post_save(self, serializer):
        if self.mode == PostingEndpoint.START:
            watch_started_thread(self.user, self.thread)
        elif self.mode == PostingEndpoint.REPLY:
            watch_replied_thread(self.user, self.thread)

        if self.mode == PostingEndpoint.REPLY:
            notify_on_new_thread_reply.delay(self.post.id)



from ..validation import filter_user_data


def test_new_user_data_is_filtered_using_default_filters(db, request):
    filtered_data = filter_user_data(
        request,
        None,
        {
            "id": "242132",
            "name": "New User",
            "email": "oauth2@example.com",
            "avatar": None,
        },
    )

    assert filtered_data == {
        "id": "242132",
        "name": "New_User",
        "email": "oauth2@example.com",
        "avatar": None,
    }


def test_existing_user_data_is_filtered_using_default_filters(user, request):
    filtered_data = filter_user_data(
        request,
        user,
        {
            "id": "242132",
            "name": user.username,
            "email": user.email,
            "avatar": None,
        },
    )

    assert filtered_data == {
        "id": "242132",
        "name": user.username,
        "email": user.email,
        "avatar": None,
    }


def test_empty_user_name_is_replaced_with_placeholder_one(db, request):
    filtered_data = filter_user_data(
        request,
        None,
        {
            "id": "242132",
            "name": None,
            "email": "oauth2@example.com",
            "avatar": None,
        },
    )

    assert filtered_data["name"].strip()


def test_missing_user_name_is_replaced_with_placeholder_one(db, request):
    filtered_data = filter_user_data(
        request,
        None,
        {
            "id": "242132",
            "name": None,
            "email": "oauth2@example.com",
            "avatar": None,
        },
    )

    assert filtered_data["name"].strip()


def test_missing_user_email_is_set_as_empty_str(db, request):
    filtered_data = filter_user_data(
        request,
        None,
        {
            "id": "242132",
            "name": "New User",
            "email": None,
            "avatar": None,
        },
    )

    assert filtered_data == {
        "id": "242132",
        "name": "New_User",
        "email": "",
        "avatar": None,
    }


def test_original_user_data_is_not_mutated_by_default_filters(user, request):
    user_data = {
        "id": "1234",
        "name": "New User",
        "email": "oauth2@example.com",
        "avatar": None,
    }

    filtered_data = filter_user_data(request, user, user_data)
    assert filtered_data == {
        "id": "1234",
        "name": "New_User",
        "email": "oauth2@example.com",
        "avatar": None,
    }

    assert user_data == {
        "id": "1234",
        "name": "New User",
        "email": "oauth2@example.com",
        "avatar": None,
    }



from django.contrib.auth import get_user_model
from django.utils.translation import npgettext
from rest_framework import serializers

from ...conf import settings

User = get_user_model()

__all__ = ["ModerateAvatarSerializer", "ModerateSignatureSerializer"]


class ModerateAvatarSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = [
            "is_avatar_locked",
            "avatar_lock_user_message",
            "avatar_lock_staff_message",
        ]


class ModerateSignatureSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = [
            "signature",
            "is_signature_locked",
            "signature_lock_user_message",
            "signature_lock_staff_message",
        ]

    def validate_signature(self, value):
        length_limit = settings.signature_length_max
        if len(value) > length_limit:
            message = npgettext(
                "signature length validator",
                "Signature can't be longer than %(limit)s character.",
                "Signature can't be longer than %(limit)s characters.",
                length_limit,
            )
            raise serializers.ValidationError(message % {"limit": length_limit})

        return value



from django.contrib.auth import update_session_auth_hash
from django.core.exceptions import PermissionDenied
from django.db import IntegrityError
from django.shortcuts import render
from django.utils.translation import pgettext

from ..credentialchange import read_new_credential
from ..decorators import deny_guests
from ..pages import usercp


@deny_guests
def index(request, *args, **kwargs):
    user_options = []
    for section in usercp.get_sections(request):
        user_options.append(
            {
                "name": str(section["name"]),
                "icon": section["icon"],
                "component": section["component"],
            }
        )

    request.frontend_context.update({"USER_OPTIONS": user_options})

    return render(request, "misago/options/noscript.html")


class ChangeError(Exception):
    pass


def confirm_change_view(f):
    @deny_guests
    def decorator(request, token):
        try:
            return f(request, token)
        except ChangeError:
            return render(request, "misago/options/credentials_error.html", status=400)

    return decorator


@confirm_change_view
def confirm_email_change(request, token):
    if request.settings.enable_oauth2_client:
        raise PermissionDenied(
            pgettext("user options", "Please use %(provider)s to change your e-mail.")
            % {"provider": request.settings.oauth2_provider}
        )

    new_credential = read_new_credential(request, "email", token)
    if not new_credential:
        raise ChangeError()

    try:
        request.user.set_email(new_credential)
        request.user.save(update_fields=["email", "email_hash"])
    except IntegrityError:
        raise ChangeError()

    message = pgettext("user options", "%(user)s, your e-mail has been changed.")
    return render(
        request,
        "misago/options/credentials_changed.html",
        {"message": message % {"user": request.user.username}},
    )


@confirm_change_view
def confirm_password_change(request, token):
    if request.settings.enable_oauth2_client:
        raise PermissionDenied(
            pgettext("user options", "Please use %(provider)s to change your password.")
            % {"provider": request.settings.oauth2_provider}
        )

    new_credential = read_new_credential(request, "password", token)
    if not new_credential:
        raise ChangeError()

    request.user.set_password(new_credential)
    update_session_auth_hash(request, request.user)
    request.user.save(update_fields=["password"])

    message = pgettext("user options", "%(user)s, your password has been changed.")
    return render(
        request,
        "misago/options/credentials_changed.html",
        {"message": message % {"user": request.user.username}},
    )



# fixme: rename this module to serialize
def hydrate_string(dry_value):
    return str(dry_value) if dry_value else ""


def dehydrate_string(wet_value):
    return str(wet_value)


def hydrate_bool(dry_value):
    return dry_value == "True"


def dehydrate_bool(wet_value):
    return "True" if wet_value else "False"


def hydrate_int(dry_value):
    return int(dry_value or 0)


def dehydrate_int(wet_value):
    return str(wet_value or 0)


def hydrate_list(dry_value):
    if dry_value:
        return [x for x in dry_value.split(",") if x]
    return []


def dehydrate_list(wet_value):
    return ",".join(wet_value) if wet_value else ""


def noop(value):
    return value


VALUE_HYDRATORS = {
    "string": (hydrate_string, dehydrate_string),
    "bool": (hydrate_bool, dehydrate_bool),
    "int": (hydrate_int, dehydrate_int),
    "list": (hydrate_list, dehydrate_list),
    "image": (noop, noop),
}


def hydrate_value(python_type, dry_value):
    try:
        value_hydrator = VALUE_HYDRATORS[python_type][0]
    except KeyError:
        raise ValueError("%s type is not hydrateable" % python_type)

    return value_hydrator(dry_value)


def dehydrate_value(python_type, wet_value):
    try:
        value_dehydrator = VALUE_HYDRATORS[python_type][1]
    except KeyError:
        raise ValueError("%s type is not dehydrateable" % python_type)

    return value_dehydrator(wet_value)



from django import forms
from django.utils.translation import npgettext_lazy, pgettext_lazy

from ....admin.forms import YesNoSwitch
from ....core.validators import validate_image_square
from ....users.validators import validate_username_content
from ... import settings
from .base import SettingsForm


class UsersSettingsForm(SettingsForm):
    settings = [
        "account_activation",
        "allow_custom_avatars",
        "avatar_upload_limit",
        "default_avatar",
        "default_gravatar_fallback",
        "blank_avatar",
        "signature_length_max",
        "subscribe_reply",
        "subscribe_start",
        "username_length_max",
        "username_length_min",
        "anonymous_username",
        "users_per_page",
        "users_per_page_orphans",
        "top_posters_ranking_length",
        "top_posters_ranking_size",
        "allow_data_downloads",
        "data_downloads_expiration",
        "allow_delete_own_account",
        "new_inactive_accounts_delete",
        "ip_storage_time",
    ]

    account_activation = forms.ChoiceField(
        label=pgettext_lazy(
            "admin users settings form", "Require new accounts activation"
        ),
        choices=[
            (
                "none",
                pgettext_lazy(
                    "admin users account activation field choice",
                    "No activation required",
                ),
            ),
            (
                "user",
                pgettext_lazy(
                    "admin users account activation field choice",
                    "Activation token sent to user e-mail",
                ),
            ),
            (
                "admin",
                pgettext_lazy(
                    "admin users account activation field choice",
                    "Activation by administrator",
                ),
            ),
            (
                "closed",
                pgettext_lazy(
                    "admin users account activation field choice",
                    "Disable new registrations",
                ),
            ),
        ],
        widget=forms.RadioSelect(),
    )
    new_inactive_accounts_delete = forms.IntegerField(
        label=pgettext_lazy(
            "admin users settings form",
            "Delete new inactive accounts if they weren't activated within this number of days",
        ),
        help_text=pgettext_lazy(
            "admin users settings form",
            "Enter 0 to never delete inactive new accounts.",
        ),
        min_value=0,
    )

    username_length_min = forms.IntegerField(
        label=pgettext_lazy(
            "admin users settings form", "Minimum allowed username length"
        ),
        min_value=2,
        max_value=20,
    )
    username_length_max = forms.IntegerField(
        label=pgettext_lazy(
            "admin users settings form", "Maximum allowed username length"
        ),
        min_value=2,
        max_value=20,
    )

    allow_custom_avatars = YesNoSwitch(
        label=pgettext_lazy("admin users settings form", "Allow custom avatar uploads"),
        help_text=pgettext_lazy(
            "admin users settings form",
            "Turning this option off will forbid forum users from uploading custom avatars.",
        ),
    )
    avatar_upload_limit = forms.IntegerField(
        label=pgettext_lazy(
            "admin users settings form", "Maximum size of uploaded avatar"
        ),
        help_text=pgettext_lazy(
            "admin users settings form",
            "Enter maximum allowed file size (in KB) for avatar uploads.",
        ),
        min_value=0,
    )
    default_avatar = forms.ChoiceField(
        label=pgettext_lazy("admin users settings form", "Default avatar"),
        choices=[
            (
                "dynamic",
                pgettext_lazy(
                    "admin users default avatar choice",
                    "Individual",
                ),
            ),
            (
                "gravatar",
                pgettext_lazy(
                    "admin users default avatar choice",
                    "Gravatar",
                ),
            ),
            (
                "gallery",
                pgettext_lazy(
                    "admin users default avatar choice",
                    "Random avatar from gallery",
                ),
            ),
        ],
        widget=forms.RadioSelect(),
    )
    default_gravatar_fallback = forms.ChoiceField(
        label=pgettext_lazy(
            "admin users settings form", "Fallback for default Gravatar"
        ),
        help_text=pgettext_lazy(
            "admin users settings form",
            "Select which avatar to use when user has no Gravatar associated with their e-mail address.",
        ),
        choices=[
            (
                "dynamic",
                pgettext_lazy("admin users gravatar fallback choice", "Individual"),
            ),
            (
                "gallery",
                pgettext_lazy(
                    "admin users gravatar fallback choice", "Random avatar from gallery"
                ),
            ),
        ],
        widget=forms.RadioSelect(),
    )
    blank_avatar = forms.ImageField(
        label=pgettext_lazy("admin users settings form", "Blank avatar"),
        help_text=pgettext_lazy(
            "admin users settings form",
            "Blank avatar is displayed in the interface when user's avatar is not available: when user was deleted or is guest. Uploaded image should be a square.",
        ),
        required=False,
    )
    blank_avatar_delete = forms.BooleanField(
        label=pgettext_lazy("admin users settings form", "Delete custom blank avatar"),
        required=False,
    )

    signature_length_max = forms.IntegerField(
        label=pgettext_lazy(
            "admin users settings form", "Maximum allowed signature length"
        ),
        min_value=10,
        max_value=5000,
    )

    subscribe_start = forms.ChoiceField(
        label=pgettext_lazy("admin users settings form", "Started threads"),
        choices=[
            ("no", pgettext_lazy("admin users settings form", "Don't watch")),
            (
                "watch",
                pgettext_lazy(
                    "admin users default subscription choice",
                    "Put on watched threads list",
                ),
            ),
            (
                "watch_email",
                pgettext_lazy(
                    "admin users default subscription choice",
                    "Put on watched threads list and e-mail user when somebody replies",
                ),
            ),
        ],
        widget=forms.RadioSelect(),
    )
    subscribe_reply = forms.ChoiceField(
        label=pgettext_lazy("admin users settings form", "Replied threads"),
        choices=[
            ("no", pgettext_lazy("admin users settings form", "Don't watch")),
            (
                "watch",
                pgettext_lazy(
                    "admin users default subscription choice",
                    "Put on watched threads list",
                ),
            ),
            (
                "watch_email",
                pgettext_lazy(
                    "admin users default subscription choice",
                    "Put on watched threads list and e-mail user when somebody replies",
                ),
            ),
        ],
        widget=forms.RadioSelect(),
    )

    users_per_page = forms.IntegerField(
        label=pgettext_lazy(
            "admin users settings form", "Number of users displayed on a single page"
        ),
        min_value=4,
    )
    users_per_page_orphans = forms.IntegerField(
        label=pgettext_lazy("admin users settings form", "Maximum orphans"),
        help_text=pgettext_lazy(
            "admin users settings form",
            "This setting prevents situations when the last page of a users list contains very few items. If number of users to be displayed on the last page is less or equal to number specified in this setting, those users will instead be appended to the previous page, reducing number of list's pages.",
        ),
        min_value=0,
    )

    top_posters_ranking_length = forms.IntegerField(
        label=pgettext_lazy(
            "admin users settings form",
            "Maximum age in days of posts that should count to the ranking position",
        ),
        min_value=1,
    )
    top_posters_ranking_size = forms.IntegerField(
        label=pgettext_lazy(
            "admin users settings form", "Maximum number of ranked users"
        ),
        min_value=2,
    )

    allow_data_downloads = YesNoSwitch(
        label=pgettext_lazy(
            "admin users settings form", "Allow users to download their data"
        )
    )
    data_downloads_expiration = forms.IntegerField(
        label=pgettext_lazy(
            "admin users settings form",
            "Maximum age in hours of data downloads before they expire",
        ),
        help_text=pgettext_lazy(
            "admin users settings form",
            "Data downloads older than specified will have their files deleted and will be marked as expired.",
        ),
        min_value=1,
    )

    allow_delete_own_account = YesNoSwitch(
        label=pgettext_lazy(
            "admin users settings form", "Allow users to delete their own accounts"
        )
    )

    ip_storage_time = forms.IntegerField(
        label=pgettext_lazy("admin users settings form", "IP storage time"),
        help_text=pgettext_lazy(
            "admin users settings form",
            "Number of days for which users IP addresses are stored in forum database. Enter zero to store registered IP addresses forever. Deleting user account always deletes the IP addresses associated with it.",
        ),
        min_value=0,
    )

    anonymous_username = forms.CharField(
        label=pgettext_lazy("admin users settings form", "Anonymous username"),
        help_text=pgettext_lazy(
            "admin users settings form",
            "This username is displayed instead of delete user's actual name next to their content.",
        ),
        min_length=1,
        max_length=15,
        validators=[validate_username_content],
    )

    def clean_blank_avatar(self):
        upload = self.cleaned_data.get("blank_avatar")
        if not upload or upload == self.initial.get("blank_avatar"):
            return None

        validate_image_square(upload.image)
        min_size = max(settings.MISAGO_AVATARS_SIZES)
        if upload.image.width < min_size:
            raise forms.ValidationError(
                npgettext_lazy(
                    "admin users settings form",
                    "Uploaded image's edge should be at least %(size)s pixel long.",
                    "Uploaded image's edge should be at least %(size)s pixels long.",
                    min_size,
                )
                % {"size": min_size}
            )

        return upload

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data.get("users_per_page_orphans") > cleaned_data.get(
            "users_per_page"
        ):
            self.add_error(
                "users_per_page_orphans",
                pgettext_lazy(
                    "admin users settings form",
                    "This value must be lower than number of users per page.",
                ),
            )
        return cleaned_data
